/**************************************************************************** 
*  Copyright 2009 C&H Technologies, Inc.  All Rights Reserved.					    * 
*****************************************************************************/
/****************************************************************************
*
*	m228_ansi.c
*
*	ANSI-C source module for the M228 Apeture A/D M-Module.  The MA209 is 
*	manufactured and supported by C&H Technologies, Inc.
*
* This file contains high level functions to control a M228 Apeture A/D
*	M-Module from C&H Technologies, Inc.  This file is written entirely in 
*	ANSI-C so that is may be compiled in any environment with a supporting 
*	ANIS-C compiler. It uses the APIS library to perform instrument 
*	communications.  
*
*	This file is organized such that modifications are easy.  All functions 
*	that might need to be modified are located at the end of this file.  
*	Comments are provided that explain tasks of the function and how it might 
*	be modified.  All other function should not need to be modified.
*
*	C&H Part Number: 11030509 
*
*	Modification History:
*
*****************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <stdarg.h>
#include "m228_ansi.h"


/**************************************************************************** 
*  Non Exported Defines
*****************************************************************************/
/*  Driver Revision  */
#define		M228_DRIVER_REV			"1.0"

//*  Other Defines  */
#define		M228_CALIBRATION_EEPROM_HEADER_STRING		"M228ApertureA/D"

/*  Calibration limits used when calibrating the peripheral variable voltage reference */
#define		M228_UNITY_CAL_LIMITS_TYPE						0						//0=unipolar positive, 1=bipolar, 2=unipolar negative 
#define		M228_UNITY_CAL_LIMITS_OFFSET_EXP			0
#define		M228_UNITY_CAL_LIMITS_OFFSET_MAX			0.002
#define		M228_UNITY_CAL_LIMITS_OFFSET_MIN			-0.002
#define		M228_UNITY_CAL_LIMITS_FULLSCALE_EXP		9.998
#define		M228_UNITY_CAL_LIMITS_FULLSCALE_MAX		9.998
#define		M228_UNITY_CAL_LIMITS_FULLSCALE_MIN		9.995


/**************************************************************************** 
*  Error Checking Macros
*****************************************************************************/
#ifndef checkParamRange
#define  checkParamRange(value, min, max, param)		if ((value > max) || (value < min)) { \
																											error = M228_ERROR_PARAMETER1 + param-1;\
																											goto Error; \
																										} else 
#endif																				
																				
#ifndef checkParamUnsignedRange
#define  checkParamUnsignedRange(value, max, param)		if (value > max) { \
																											error = M228_ERROR_PARAMETER1 + param-1;\
																											goto Error; \
																										} else 
#endif																				

#ifndef checkParamPointer
#define  checkParamPointer(pointer, param)		if (pointer == NULL) { \
																							error = M228_ERROR_PARAMETER1 + param-1;\
																							goto Error; \
																							} else 
#endif																				

																				
#ifndef checkErr
#define checkErr(fCall)      if (error = (fCall), (error < 0)) \
                                 {goto Error;}  else
#endif

#ifndef apisCheckErr
#define apisCheckErr(fCall)      if (error = (fCall), (error = ((error != 0) ? error : M228_SUCCESS))) \
                                 {	handle->apis_error = error;\
																		error = M228_ERROR_APIS; \
																		goto Error;}  else
#endif

#ifndef apisCheckErrInit
#define apisCheckErrInit(fCall)      if (error = (fCall), (error = (error != 0) ? error : M228_SUCCESS)) \
                                 {	(*handle)->apis_error = error;\
																		error = M228_ERROR_APIS; \
																		goto Error;}  else
#endif


/**************************************************************************** 
*  Defined Structures
*****************************************************************************
/*  None  */

/**************************************************************************** 
*  Constants
*****************************************************************************/

static const char *errorMessageTable[M228_SPECIFIC_ERROR_END - M228_SPECIFIC_ERROR_BASE] = 
						{	"Fifo empty. No valid data was found in the fifo.",
							"An invalid parameter was passed to the function in parameter 1",
							"An invalid parameter was passed to the function in parameter 2",
							"An invalid parameter was passed to the function in parameter 3",
							"An invalid parameter was passed to the function in parameter 4",
							"An invalid parameter was passed to the function in parameter 5",
							"An invalid parameter was passed to the function in parameter 6",
							"An invalid parameter was passed to the function in parameter 7",
							"An invalid parameter was passed to the function in parameter 8",
							"An invalid parameter was passed to the function in parameter 9",
							"An invalid parameter was passed to the function in parameter 10",
							"The module returned an invalid ID.",
							"An error occurred in the APIS Platform Support Module.  Use the function ma204_GetApisError() to retrieve the APIS error code.",
							"The operation could not be completed because the system is out of memory.",
							"I2C Timeout Occured.",
							"Calibration value could not be found while calibrating to internal references.",
							"Calibration was aborted by the calling application."
						};

static const char *warningMessageTable[M228_SPECIFIC_WARNING_END - M228_SPECIFIC_WARNING_BASE] =
						{	"There were less samples stored in the FIFO than what was requested.",
							"The calibration entry is empty.",
							"The A/D value indicates an over voltage condition.",
							"The A/D value indicates an under voltage condition."
						};

static const int fgain_multiplier[4] = { 1, 2, 5, 10}; 
static const int bgain_multiplier[8] = { 1, 2, 5, 10, 20, 50, 100, 0}; 

/**************************************************************************** 
*  Non Exported Functions
*****************************************************************************/
static Int32 m228_IDWriteByte(M228_HANDLE handle, UInt8 value);
static Int32 m228_IDWriteBit(M228_HANDLE handle,  Bool value);
static Int32 m228_IDReadByte(M228_HANDLE handle, UInt8 *value);
static Int32 m228_IDReadBit(M228_HANDLE handle,  Bool *value);
static Int32 m228_ReadTimeValuePair(M228_HANDLE handle, Int32 interface, Bool width32bit, UInt32 *timeStamp, Real64 *value, UInt16 *rawData);
static void m228_CalMessage(void (*user_CalMessage)(Int8 *), Int8 *msg, ...);
static Int32 m228_FindCalibrationValue(M228_HANDLE handle, Bool type, Real64 exp, Real64 max, Real64 min, UInt8 *value, Bool *abort, void (*user_CalMessage)(Int8 *));
static Int32 m228_SetAdjustedPeripheralVoltage(M228_HANDLE handle, UInt8 offset_cal_val, UInt8 fullscale_cal_val, Real64 voltage);
static Int32 m228_PerformUnityCalibration(M228_HANDLE handle, UInt8 *offset_val, UInt8 *fullscale_val, M228_CALIBRATION_LIMITS *limits, Bool *abort, void (*user_CalMessage)(Int8 *));



/*****************************************************************************
*	m228_QueryModuleIdentification 
*
*	Returns the Model, Configuration, and Logic Major & Minor Revision level
*	
******************************************************************************/
Int32 m228_QueryModuleIdentification (M228_HANDLE handle, UInt8 *model, UInt8 *config, UInt8 *log_major_rev, UInt8 *log_minor_rev)
{	Int32 error = M228_SUCCESS;
	UInt16 value;
	
/*  Check parameters  */
	checkParamPointer(model, 2);
	checkParamPointer(config, 3);
	checkParamPointer(log_major_rev, 4);
	checkParamPointer(log_minor_rev, 5);
	
/*  Read the ID register  */
	checkErr(m228_ReadRegister16(handle, M228_ID_REGISTER, &value));

	*model = (value & M228_MODEL_BIT_MASK) >> M228_MODEL_BIT_NUM;
	*config	= (value & M228_CONFIG_BIT_MASK) >> M228_CONFIG_BIT_NUM;

/*  Read the Revision register  */
	checkErr(m228_ReadRegister16(handle, M228_REVISION_REGISTER, &value));

	*log_major_rev = (value & M228_LOGIC_MAJOR_REV_BIT_MASK) >> M228_LOGIC_MAJOR_REV_BIT_NUM;
	*log_minor_rev = (value & M228_LOGIC_MINOR_REV_BIT_MASK) >> M228_LOGIC_MINOR_REV_BIT_NUM;
	
Error:
	return error;
}


/*****************************************************************************
*	m228_QueryFifoStatus 
*
*	Returns the values of the three fifo status bits.  These bits are in
*	the control status register bits 13-15.
******************************************************************************/
Int32 m228_QueryFifoStatus (M228_HANDLE handle, Bool *fifoFull, Bool *fifoHalfFull, Bool *fifoQtrFull, Bool *fifoNotEmpty)
{	Int32 error = M228_SUCCESS;
	UInt16 value;
	
/*  Check parameters  */
	checkParamPointer(fifoFull, 2);
	checkParamPointer(fifoHalfFull, 3);
	checkParamPointer(fifoQtrFull, 4);
	checkParamPointer(fifoNotEmpty, 5);
	
	*fifoNotEmpty = M228_FALSE;
	*fifoFull = M228_FALSE;
	*fifoHalfFull = M228_FALSE;
	*fifoQtrFull = M228_FALSE;
	
/*  Read the control/status register  */
	checkErr(m228_ReadRegister16(handle, M228_MASTER_CONTROL_REGISTER, &value));
	
/*  Check each status bit  */	
	if(value & M228_FIFO_NOT_EMPTY_BIT_MASK)
		*fifoNotEmpty = M228_TRUE;
	if(value & M228_FIFO_FULL_BIT_MASK)
		*fifoFull = M228_TRUE;
	if(value & M228_FIFO_HALF_FULL_BIT_MASK)
		*fifoHalfFull = M228_TRUE;
	if(value & M228_FIFO_QTR_FULL_BIT_MASK)
		*fifoQtrFull = M228_TRUE;

Error:
	return error;
}


/*****************************************************************************
*	m228_FifoUnreadCount 
*
*	Returns the value of the fifo unread count register
******************************************************************************/
Int32 m228_FifoUnreadCount (M228_HANDLE handle, UInt32 *unreadCount)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamPointer(unreadCount, 2);

	checkErr(m228_ReadRegister16(handle, M228_FIFO_UNREAD_COUNT_REGISTER_HIGH, &data));
	*unreadCount = (data << 16) & 0xFFFF0000;
	checkErr(m228_ReadRegister16(handle, M228_FIFO_UNREAD_COUNT_REGISTER_LOW, &data));
	*unreadCount |= data;
Error:
	return error;
}


/*****************************************************************************
*	m228_ResetFifo
*
*	Resets the fifo.  This will also clear any of the fifo interrupts pending
******************************************************************************/
Int32 m228_ResetFifo (M228_HANDLE handle)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Read the current value of the control/status register  */
	checkErr(m228_ReadRegister16(handle, M228_MASTER_CONTROL_REGISTER, &data));
	
/*  Set the reset fifo bit and write to the register  */	
	checkErr(m228_WriteRegister16(handle, M228_MASTER_CONTROL_REGISTER, (data | M228_RESET_FIFO_BIT_MASK)));
Error:
	return error;
}


/*****************************************************************************
*	m228_ResetTimeStamp 
*
*	Resets the time stamp.
******************************************************************************/
Int32 m228_ResetTimeStamp (M228_HANDLE handle)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Read the current value of the control/status register  */
	checkErr(m228_ReadRegister16(handle, M228_MASTER_CONTROL_REGISTER, &data));

/*  Set time stamp reset bit and clear the fifo status bits  */
	data &= ~M228_STATUS_BIT_MASK;					//do not write status bits to avoid clearing irq pendings
	data |= M228_RESET_TIME_STAMP_BIT_MASK;
	
/*  Write the new value to the register  */	
	checkErr(m228_WriteRegister16(handle, M228_MASTER_CONTROL_REGISTER, data));
Error:
	return error;
}


/*****************************************************************************
*	m228_QueryTimestamp 
*
******************************************************************************/
Int32 m228_QueryTimestamp (M228_HANDLE handle, UInt32 *timestamp)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamPointer(timestamp, 2);

	checkErr(m228_ReadRegister16(handle, M228_TIMESTAMP_REGISTER_HIGH, &data));
	*timestamp = (data << 16) & 0xFFFF0000;
	checkErr(m228_ReadRegister16(handle, M228_TIMESTAMP_REGISTER_LOW, &data));
	*timestamp |= data;
Error:
	return error;
}


/*****************************************************************************
*	m228_QueryTimestampRollover 
*
******************************************************************************/
Int32 m228_QueryTimestampRollover (M228_HANDLE handle, Bool *ts_rollover)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamPointer(ts_rollover, 2);

	checkErr(m228_ReadRegister16(handle, M228_MASTER_CONTROL_REGISTER, &data));
	if(data & M228_TSR_BIT_MASK)
		*ts_rollover = M228_TRUE;
	else
		*ts_rollover = M228_FALSE;
Error:
	return error;
}


/*****************************************************************************
*	m228_LastStoredValue
*
*	Returns the value of the last value stored register
******************************************************************************/
Int32 m228_LastStoredValue (M228_HANDLE handle, Real64 *last_stored, UInt16 *last_stored_raw)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

	checkErr(m228_ReadRegister16(handle, M228_LAST_VALUE_STORED_REGISTER, &data));
	
	if(last_stored_raw != NULL)
		*last_stored_raw = data;

	if(last_stored != NULL)
/**************
Add real value calculation
***************/
		*last_stored = 0.0;      

Error:
	return error;
}


/*****************************************************************************
*	m228_ConfigureSampleClock
*
******************************************************************************/
Int32 m228_ConfigureSampleClock (M228_HANDLE handle, UInt8 source, UInt8 prescaler)
{	Int32 error = M228_SUCCESS;
	UInt16 data = 0;

/*  Check parameters  */
	checkParamRange(source, 0, 7, 2);
	checkParamRange(prescaler, 0, 15, 3);

	checkErr(m228_ReadRegister16(handle, M228_CLOCK_APERTURE_CONTROL_REGISTER, &data));

/*  Form new register value and write it  */
	data &= ~(M228_CLKSRC_BIT_MASK | M228_PSC_BIT_MASK);	
	data |= ((source << M228_CLKSRC_BIT_NUM) | (prescaler << M228_PSC_BIT_NUM)); 									 
																																							 
	checkErr(m228_WriteRegister16(handle, M228_CLOCK_APERTURE_CONTROL_REGISTER, data));

	if((source == 0) && (prescaler == 0))
		checkErr(m228_ConfigureWarpMode(handle, 1));
	else if (source == 0)
		checkErr(m228_ConfigureWarpMode(handle, 0));
	//else leave warp mode as is
	
	
Error:
	return error;
}


/*****************************************************************************
*	m228_QuerySampleClockConfiguration
*
******************************************************************************/
Int32 m228_QuerySampleClockConfiguration (M228_HANDLE handle, UInt8 *source, UInt8 *prescaler)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamPointer(source, 2);
	checkParamPointer(prescaler, 3);

	checkErr(m228_ReadRegister16(handle, M228_CLOCK_APERTURE_CONTROL_REGISTER, &data));
	*source = (data & M228_CLKSRC_BIT_MASK) >> M228_CLKSRC_BIT_NUM;
	*prescaler = (data & M228_PSC_BIT_MASK) >> M228_PSC_BIT_NUM;	
	

Error:
	return error;
}


/*****************************************************************************
*	m228_ConfigureWarpMode
*
******************************************************************************/
Int32 m228_ConfigureWarpMode (M228_HANDLE handle, Bool warp_mode)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamRange(warp_mode, 0, 1, 2);

	checkErr(m228_ReadRegister16(handle, M228_ANALOG_INPUT_CONTROL_REGISTER, &data));
	if(warp_mode == M228_TRUE)
		data |= M228_WARP_BIT_MASK;
	else
		data &= ~M228_WARP_BIT_MASK;
	checkErr(m228_WriteRegister16(handle, M228_ANALOG_INPUT_CONTROL_REGISTER, data));

Error:
	return error;
}


/*****************************************************************************
*	chm228_QueryWarpMode
*
******************************************************************************/
Int32 m228_QueryWarpMode (M228_HANDLE handle, Bool *warp_mode)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamPointer(warp_mode, 2);

	checkErr(m228_ReadRegister16(handle, M228_ANALOG_INPUT_CONTROL_REGISTER, &data));
	if(data & M228_WARP_BIT_MASK)
		*warp_mode = M228_TRUE;
	else
		*warp_mode = M228_FALSE;
	
Error:
	return error;
}


/*****************************************************************************
*	m228_QueryFilterConfiguration
*
******************************************************************************/
Int32 m228_QueryFilterConfiguration (M228_HANDLE handle, UInt8 *cutoff, UInt8 *select)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamPointer(cutoff, 2);
	checkParamPointer(select, 3);
	
	checkErr(m228_ReadRegister16(handle, M228_FILTER_CONTROL_REGISTER, &data));
	*cutoff = (data & M228_FCUTOFF_BIT_MASK) >> M228_FCUTOFF_BIT_NUM;
	*select = (data & M228_FSEL_BIT_MASK) >> M228_FSEL_BIT_NUM;

Error:
	return error;
}


/*****************************************************************************
*	m228_ConfigureFilter
*
******************************************************************************/
Int32 m228_ConfigureFilter (M228_HANDLE handle, UInt8 cutoff, UInt8 select)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamRange(select, 0, 2, 3);
	if(select == M228_FILTER_ELLIPTIC)   //elliptic
		checkParamRange(cutoff, 0, 6, 2);
	else if(select == M228_FILTER_BESSEL)
		checkParamRange(cutoff, 0, 8, 2);
	
	checkErr(m228_ReadRegister16(handle, M228_FILTER_CONTROL_REGISTER, &data));
	data &= ~(M228_FCUTOFF_BIT_MASK | M228_FSEL_BIT_MASK);
	data |= (cutoff << M228_FCUTOFF_BIT_NUM) | (select << M228_FSEL_BIT_NUM); 
	checkErr(m228_WriteRegister16(handle, M228_FILTER_CONTROL_REGISTER, data));

Error:
	return error;
}


/*****************************************************************************
*	m228_ConfigureApertureMode
*
******************************************************************************/
Int32 m228_ConfigureApertureMode (M228_HANDLE handle, UInt8 aperture_mode)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamRange(aperture_mode, 0, 6, 2);
	
	checkErr(m228_ReadRegister16(handle, M228_CLOCK_APERTURE_CONTROL_REGISTER, &data));
	data &= ~M228_APMODE_BIT_MASK;
	data |= aperture_mode << M228_APMODE_BIT_NUM;
	checkErr(m228_WriteRegister16(handle, M228_CLOCK_APERTURE_CONTROL_REGISTER, data));

Error:
	return error;
}


/*****************************************************************************
*	m228_QueryApertureMode
*
******************************************************************************/
Int32 m228_QueryApertureMode (M228_HANDLE handle, UInt8 *aperture_mode)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamPointer(aperture_mode, 2);
	
	checkErr(m228_ReadRegister16(handle, M228_CLOCK_APERTURE_CONTROL_REGISTER, &data));
	*aperture_mode = (data & M228_APMODE_BIT_MASK) >> M228_APMODE_BIT_NUM;

Error:
	return error;
}


/*****************************************************************************
*	m228_ConfigureAnalogFrontEnd
*
******************************************************************************/
Int32 m228_ConfigureAnalogFrontEnd (M228_HANDLE handle, Bool divider, UInt8 front_end_gain, UInt8 back_end_gain)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamRange(divider, 0, 1, 2);
	checkParamRange(front_end_gain, 0, 3, 3);
	checkParamRange(back_end_gain, 0, 7, 4);

	checkErr(m228_ReadRegister16(handle, M228_ANALOG_INPUT_CONTROL_REGISTER, &data));
	data &= ~(M228_IVDD_BIT_MASK | M228_FGAIN_BIT_MASK | M228_BGAIN_BIT_MASK);
	data |= (divider << M228_IVDD_BIT_NUM) | (front_end_gain << M228_FGAIN_BIT_NUM) | (back_end_gain <<M228_BGAIN_BIT_NUM);
	checkErr(m228_WriteRegister16(handle, M228_ANALOG_INPUT_CONTROL_REGISTER, data));
	
	
Error:
	return error;
}


/*****************************************************************************
*	m228_QueryAnalogFrontEndConfiguration
*
******************************************************************************/
Int32 m228_QueryAnalogFrontEndConfiguration (M228_HANDLE handle, Bool *divider, UInt8 *front_end_gain, UInt8 *back_end_gain)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

	checkErr(m228_ReadRegister16(handle, M228_ANALOG_INPUT_CONTROL_REGISTER, &data));

	if(divider != NULL)
		*divider = (data & M228_IVDD_BIT_MASK) >> M228_IVDD_BIT_NUM;
	
	if(front_end_gain != NULL)
		*front_end_gain = (data & M228_FGAIN_BIT_MASK) >> M228_FGAIN_BIT_NUM;

	if(back_end_gain != NULL)
		*back_end_gain = (data & M228_BGAIN_BIT_MASK) >> M228_BGAIN_BIT_NUM; 
	
Error:
	return error;
}


/*****************************************************************************
*	m228_SimulationMode
*
******************************************************************************/
Int32 m228_SimulationMode (M228_HANDLE handle, UInt8 mode, Bool factory_test)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamRange(mode, 0, 2, 2);
	checkParamRange(factory_test, 0, 1, 3);

	checkErr(m228_ReadRegister16(handle, M228_ANALOG_INPUT_CONTROL_REGISTER, &data));
	data &= ~(M228_SIM_BIT_MASK);
	data &= ~(M228_FT_BIT_MASK);
	data |= (mode << M228_SIM_BIT_NUM);
	if(factory_test)
		data |= (1 << M228_FT_BIT_NUM);
	checkErr(m228_WriteRegister16(handle, M228_ANALOG_INPUT_CONTROL_REGISTER, data));
	
Error:
	return error;
}


/*****************************************************************************
*	m228_QuerySimulationMode
*
******************************************************************************/
Int32 m228_QuerySimulationMode (M228_HANDLE handle, UInt8 *mode, Bool *factory_test)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamPointer(mode, 2);
	checkParamPointer(factory_test, 3);

	checkErr(m228_ReadRegister16(handle, M228_ANALOG_INPUT_CONTROL_REGISTER, &data));
	*mode = (data & M228_SIM_BIT_MASK) >> M228_SIM_BIT_NUM;
	if((data & M228_FT_BIT_MASK) == 0)
		*factory_test = 0;
	else
		*factory_test = 1;
	
Error:
	return error;
}


/*****************************************************************************
*	m228_ConfigureReferenceInjection
*
******************************************************************************/
Int32 m228_ConfigureReferenceInjection (M228_HANDLE handle, Bool ref_enable, UInt8 ref_select)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamRange(ref_enable, 0,1,2);
	checkParamRange(ref_select, 0,7,3);

	checkErr(m228_ReadRegister16(handle, M228_ANALOG_INPUT_CONTROL_REGISTER, &data));
	data &= ~(M228_REFE_BIT_MASK | M228_REFSEL_BIT_MASK);
	data |= (ref_enable << M228_REFE_BIT_NUM) | (ref_select << M228_REFSEL_BIT_NUM);
	checkErr(m228_WriteRegister16(handle, M228_ANALOG_INPUT_CONTROL_REGISTER, data));
	
	
Error:
	return error;
}

 
/*****************************************************************************
*	m228_QueryReferenceInjection
*
******************************************************************************/
Int32 m228_QueryReferenceInjection (M228_HANDLE handle, Bool *ref_enable, UInt8 *ref_select)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

	checkErr(m228_ReadRegister16(handle, M228_ANALOG_INPUT_CONTROL_REGISTER, &data));

	if(ref_enable != NULL)
		*ref_enable = (data & M228_REFE_BIT_MASK) >> M228_REFE_BIT_NUM;
	
	if(ref_select != NULL)
		*ref_select = (data & M228_REFSEL_BIT_MASK) >> M228_REFSEL_BIT_NUM;

Error:
	return error;
}


/*****************************************************************************
*	m228_CurrentValue
*
******************************************************************************/
Int32 m228_CurrentValue (M228_HANDLE handle, Real64 *real_value, UInt16 *raw_value)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

	checkErr(m228_ReadRegister16(handle, M228_CURRENT_VALUE_REGISTER, &data));
	
	if(raw_value != NULL)
		*raw_value = data;

	if(real_value != NULL)
		checkErr(m228_ConvertToVoltage(handle, &data, real_value, 1, 1, 0, 0, 0));  //ConvertToVoltage checks for over and under voltage
	else
	{	if(data == 0x2000)	
			error = M228_WARNING_UNDERVOLTAGE;	   
		else if(data == 0x1FFF)
			error = M228_WARNING_OVERVOLTAGE;
	}
	
Error:
	return error;
}


/*****************************************************************************
*	m228_SetPeripheralVoltage
*
******************************************************************************/
Int32 m228_SetPeripheralVoltage (M228_HANDLE handle, Real64 voltage)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamRange(voltage, -10.0, 10.0, 2);
	
	data = (UInt16)(((-voltage + M228_PVARV_OFFSET) / M228_PVARV_RESOLUTION) + 0.5);
	checkErr(m228_WriteRegister16(handle, M228_PERIPHERAL_VAR_VOLTAGE_REGISTER, data));
Error:
	return error;
}


/*****************************************************************************
*	m228_QueryPeripheralVoltage
*
******************************************************************************/
Int32 m228_QueryPeripheralVoltage (M228_HANDLE handle, Real64 *voltage)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamPointer(voltage, 2);
	
	checkErr(m228_ReadRegister16(handle, M228_PERIPHERAL_VAR_VOLTAGE_REGISTER, &data));
	*voltage = (((Real64)data * M228_PVARV_RESOLUTION) - M228_PVARV_OFFSET) * -1;
Error:
	return error;
}


/*****************************************************************************
*	m228_ApertureSelect
*
******************************************************************************/
Int32 m228_ApertureSelect (M228_HANDLE handle, Bool aperture_select)
{	Int32 error = M228_SUCCESS;
	UInt16 data;
	
/*  Check parameters  */
	checkParamRange(aperture_select, 0, 1, 2);

	checkErr(m228_ReadRegister16(handle, M228_MASTER_CONTROL_REGISTER, &data));
	if(aperture_select == M228_TRUE)
		data |= M228_ASEL_BIT_MASK;
	else
		data &= ~M228_ASEL_BIT_MASK;
	checkErr(m228_WriteRegister16(handle, M228_MASTER_CONTROL_REGISTER, data));
Error:
	return error;
}


/*****************************************************************************
*	m228_QueryApertureSelect
*
******************************************************************************/
Int32 m228_QueryApertureSelect (M228_HANDLE handle, Bool *aperture_select)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamPointer(aperture_select, 2);

	checkErr(m228_ReadRegister16(handle, M228_MASTER_CONTROL_REGISTER, &data));
	if(data & M228_ASEL_BIT_MASK)
		*aperture_select = M228_TRUE;
	else
		*aperture_select = M228_FALSE;
Error:
	return error;
}


/*****************************************************************************
*	m228_SetRawAperture
*
******************************************************************************/
Int32 m228_SetRawAperture (M228_HANDLE handle, Bool aperture_select, UInt16 aperture_high, UInt16 aperture_low)
{	Int32 error = M228_SUCCESS;

/*  Check parameters  */
	checkParamRange(aperture_select, 0, 1, 2);
	checkParamRange(aperture_high, 0, 0x3FFF, 3);
	checkParamRange(aperture_low, 0, 0x3FFF, 4);

	if(aperture_select == M228_APERTURE_SELECT_A)
	{	checkErr(m228_WriteRegister16(handle, M228_APERTURE_HIGH_A_REGISTER, aperture_high));
		checkErr(m228_WriteRegister16(handle, M228_APERTURE_LOW_A_REGISTER, aperture_low));
	}
	else
	{	checkErr(m228_WriteRegister16(handle, M228_APERTURE_HIGH_B_REGISTER, aperture_high));
		checkErr(m228_WriteRegister16(handle, M228_APERTURE_LOW_B_REGISTER, aperture_low));
	}
		
Error:
	return error;
}


/*****************************************************************************
*	m228_QueryAperture
*
******************************************************************************/
Int32 m228_QueryAperture (M228_HANDLE handle, Bool aperture_select, UInt16 *aperture_high_raw, UInt16 *aperture_low_raw, Real64 *aperture_high, Real64 *aperture_low)
{	Int32 error = M228_SUCCESS;
	UInt16 data_high, data_low;

/*  Check parameters  */
	checkParamRange(aperture_select, 0, 1, 2);

	if(aperture_select == M228_APERTURE_SELECT_A)
	{	checkErr(m228_ReadRegister16(handle, M228_APERTURE_HIGH_A_REGISTER, &data_high));
		checkErr(m228_ReadRegister16(handle, M228_APERTURE_LOW_A_REGISTER, &data_low));
	}
	else
	{	checkErr(m228_ReadRegister16(handle, M228_APERTURE_HIGH_B_REGISTER, &data_high));
		checkErr(m228_ReadRegister16(handle, M228_APERTURE_LOW_B_REGISTER, &data_low));
	}
	if(aperture_high_raw != NULL)
		*aperture_high_raw = data_high;
	if(aperture_low_raw != NULL)
		*aperture_low_raw = data_low;
	if(aperture_high != NULL)
		*aperture_high = 0;
	if(aperture_low != NULL)
		*aperture_low = 0;
	
		
Error:
	return error;
}


/*****************************************************************************
*	m228_ResetTimestampRollover
*
******************************************************************************/
Int32 m228_ResetTimestampRollover (M228_HANDLE handle)
{	Int32 error = M228_SUCCESS;
	UInt16 data;					
	
	checkErr(m228_ReadRegister16(handle, M228_MASTER_CONTROL_REGISTER, &data));
	
	data &= ~M228_STATUS_BIT_MASK;			/* 	Do not write to status bits ss we don't clear irq pendings  */																																							
	
	checkErr(m228_WriteRegister16(handle, M228_MASTER_CONTROL_REGISTER, data | M228_TSR_BIT_MASK));

Error:
	return error;
}


/*****************************************************************************
*	m228_SoftwareClock
*
******************************************************************************/
Int32 m228_SoftwareClock (M228_HANDLE handle)
{	Int32 error = M228_SUCCESS;
	UInt16 data;					
	
	checkErr(m228_ReadRegister16(handle, M228_CLOCK_APERTURE_CONTROL_REGISTER, &data));
	checkErr(m228_WriteRegister16(handle, M228_CLOCK_APERTURE_CONTROL_REGISTER, data | M228_SOFTWARE_CLK_BIT_MASK));

Error:
	return error;
}


/*****************************************************************************
*	m228_Run
*
******************************************************************************/
Int32 m228_Run (M228_HANDLE handle, Bool ts_run, Bool conv_enable)
{	Int32 error = M228_SUCCESS;
	UInt16 data;					
	
/*  Check parameters  */
	checkParamRange(ts_run, 0, 1, 2);
	checkParamRange(conv_enable, 0, 1, 3);
	
	checkErr(m228_ReadRegister16(handle, M228_MASTER_CONTROL_REGISTER, &data));
	
	data &= ~M228_STATUS_BIT_MASK;			/* 	Do not write to status bits ss we don't clear irq pendings  */																																							
	
	if(ts_run == M228_TRUE)
		data |= M228_TRUN_BIT_MASK;
	else
		data &= ~M228_TRUN_BIT_MASK;

	if(conv_enable == M228_TRUE)
		data |= M228_ACE_BIT_MASK;
	else
		data &= ~M228_ACE_BIT_MASK;
	
	checkErr(m228_WriteRegister16(handle, M228_MASTER_CONTROL_REGISTER, data));

Error:
	return error;
}


/*****************************************************************************
*	m228_QueryRun
*
******************************************************************************/
Int32 m228_QueryRun (M228_HANDLE handle, Bool *ts_run, Bool *conv_enable)
{	Int32 error = M228_SUCCESS;
	UInt16 data;
	
	checkErr(m228_ReadRegister16(handle, M228_MASTER_CONTROL_REGISTER, &data));

	if(ts_run != NULL)
	{	if(data & M228_TRUN_BIT_MASK)
			*ts_run = M228_TRUE;
		else
			*ts_run = M228_FALSE;
	}

	if(conv_enable != NULL)
	{	if(data & M228_ACE_BIT_MASK)
			*conv_enable = M228_TRUE;
		else
			*conv_enable = M228_FALSE;
	}
	
Error:
	return error;
}


/*****************************************************************************
*	m228_ForceStore
*
******************************************************************************/
Int32 m228_ForceStore (M228_HANDLE handle, Bool force_store)
{	Int32 error = M228_SUCCESS;
	UInt16 data;					
	
/*  Check parameters  */
	checkParamRange(force_store, 0, 1, 2);
	
	checkErr(m228_ReadRegister16(handle, M228_MASTER_CONTROL_REGISTER, &data));
	
	data &= ~M228_STATUS_BIT_MASK;			/* 	Do not write to status bits ss we don't clear irq pendings  */																																							
	
	if(force_store == M228_TRUE)
		data |= M228_FST_BIT_MASK;
	else
		data &= ~M228_FST_BIT_MASK;

	checkErr(m228_WriteRegister16(handle, M228_MASTER_CONTROL_REGISTER, data));

Error:
	return error;
}


/*****************************************************************************
*	m228_QueryForceStore
*
******************************************************************************/
Int32 m228_QueryForceStore (M228_HANDLE handle, Bool *force_store)
{	Int32 error = M228_SUCCESS;
	UInt16 data;
	
/*  Check parameters  */
	checkParamPointer(force_store, 2);
	
	checkErr(m228_ReadRegister16(handle, M228_MASTER_CONTROL_REGISTER, &data));

	if(data & M228_FST_BIT_MASK)
		*force_store = M228_TRUE;
	else
		*force_store = M228_FALSE;
	
Error:
	return error;
}


/*****************************************************************************
*	m228_DataStorageMode
*
******************************************************************************/
Int32 m228_DataStorageMode (M228_HANDLE handle, UInt8 storage_mode)
{	Int32 error = M228_SUCCESS;
	UInt16 data;					
	
/*  Check parameters  */
	checkParamRange(storage_mode, 0, 3, 2);
	
	checkErr(m228_ReadRegister16(handle, M228_MASTER_CONTROL_REGISTER, &data));
	
	data &= ~M228_STATUS_BIT_MASK;			/* 	Do not write to status bits so we don't clear irq pendings  */																																							

	data &= ~M228_STORAGE_MODE_BIT_MASK;
	data |= storage_mode << M228_STORAGE_MODE_BIT_NUM;

	checkErr(m228_WriteRegister16(handle, M228_MASTER_CONTROL_REGISTER, data));

Error:
	return error;
}


/*****************************************************************************
*	m228_QueryDataStorageMode
*
******************************************************************************/
Int32 m228_QueryDataStorageMode (M228_HANDLE handle, UInt8 *storage_mode)
{	Int32 error = M228_SUCCESS;
	UInt16 data;
	
/*  Check parameters  */
	checkParamPointer(storage_mode, 2);
	
	checkErr(m228_ReadRegister16(handle, M228_MASTER_CONTROL_REGISTER, &data));

	*storage_mode = (data & M228_STORAGE_MODE_BIT_MASK) >> M228_STORAGE_MODE_BIT_NUM;
	
Error:
	return error;
}


/*****************************************************************************
*	m228_DataCaptureMode
*
******************************************************************************/
Int32 m228_DataCaptureMode (M228_HANDLE handle, Bool continuous_capture)
{	Int32 error = M228_SUCCESS;
	UInt16 data;					
	
/*  Check parameters  */
	checkParamRange(continuous_capture, 0, 1, 2);
	
	checkErr(m228_ReadRegister16(handle, M228_MASTER_CONTROL_REGISTER, &data));
	
	data &= ~M228_STATUS_BIT_MASK;			/* 	Do not write to status bits ss we don't clear irq pendings  */																																							
	
	if(continuous_capture == M228_TRUE)
		data |= M228_CCM_BIT_MASK;
	else
		data &= ~M228_CCM_BIT_MASK;

	checkErr(m228_WriteRegister16(handle, M228_MASTER_CONTROL_REGISTER, data));

Error:
	return error;
}


/*****************************************************************************
*	m228_QueryDataCaptureMode
*
******************************************************************************/
Int32 m228_QueryDataCaptureMode (M228_HANDLE handle, Bool *continuous_capture)
{	Int32 error = M228_SUCCESS;
	UInt16 data;
	
/*  Check parameters  */
	checkParamPointer(continuous_capture, 2);
	
	checkErr(m228_ReadRegister16(handle, M228_MASTER_CONTROL_REGISTER, &data));

	if(data & M228_CCM_BIT_MASK)
		*continuous_capture = M228_TRUE;
	else
		*continuous_capture = M228_FALSE;
	
Error:
	return error;
}


/*****************************************************************************
*	m228_CalibrationOffsetAdjust
*
******************************************************************************/
Int32 m228_CalibrationOffsetAdjust (M228_HANDLE handle, UInt8 offset_adj)
{	Int32 error = M228_SUCCESS;
	UInt16 data;					
	UInt8 data8;
	
/*  Check parameters  */
	checkParamRange(offset_adj, 0, 0xFF, 2);
	
	checkErr(m228_QueryCalibrationFullScaleAdjust (handle, &data8));
	
	data = (data8 << M228_FSVAL_BIT_NUM) | (offset_adj << M228_OSVAL_BIT_NUM);	
	checkErr(m228_WriteRegister16(handle, M228_CAL_FULLSCALE_OFFSET_REGISTER, data));

Error:
	return error;
}


/*****************************************************************************
*	m228_QueryCalibrationOffsetAdjust
*
******************************************************************************/
Int32 m228_QueryCalibrationOffsetAdjust (M228_HANDLE handle, UInt8 *offset_adj)
{	Int32 error = M228_SUCCESS;
	UInt16 data;
	Real64 start_time, elasped_time;
	
/*  Check parameters  */
	checkParamPointer(offset_adj, 2);
	
	checkErr(m228_ReadRegister16(handle, M228_CAL_EEPROM_CONTROL_REGISTER, &data));
	
	elasped_time = 0;
	start_time = (Real64)clock() / (Real64)CLOCKS_PER_SEC;
	while (((data & M228_I2C_OPC_BIT_MASK) == 0) && (elasped_time < 0.5))
	{
		checkErr(m228_ReadRegister16(handle, M228_CAL_EEPROM_CONTROL_REGISTER, &data));
		elasped_time = ((Real64)clock() / (Real64)CLOCKS_PER_SEC) - start_time;
	}
	if (elasped_time < 0.5)
	{	checkErr(m228_ReadRegister16(handle, M228_CAL_FULLSCALE_OFFSET_REGISTER, &data));
		*offset_adj = (data & M228_OSVAL_BIT_MASK) >> M228_OSVAL_BIT_NUM; 
	} 
	else
		error = M228_ERROR_I2C_TIMEOUT;
	
Error:
	return error;
}


/*****************************************************************************
*	m228_CalibrationFullScaleAdjust
*
******************************************************************************/
Int32 m228_CalibrationFullScaleAdjust (M228_HANDLE handle, UInt8 full_scale_adj)
{	Int32 error = M228_SUCCESS;
	UInt16 data;					
	UInt8 data8;
	
/*  Check parameters  */
	checkParamRange(full_scale_adj, 0, 0xFF, 2);

	checkErr(m228_QueryCalibrationOffsetAdjust (handle, &data8));
	
	data = (data8 << M228_OSVAL_BIT_NUM) | (full_scale_adj << M228_FSVAL_BIT_NUM);	
	checkErr(m228_WriteRegister16(handle, M228_CAL_FULLSCALE_OFFSET_REGISTER, data));

Error:
	return error;
}


/*****************************************************************************
*	m228_QueryCalibrationFullScaleAdjust
*
******************************************************************************/
Int32 m228_QueryCalibrationFullScaleAdjust (M228_HANDLE handle, UInt8 *full_scale_adj)
{	Int32 error = M228_SUCCESS;
	UInt16 data;
	Real64 start_time, elasped_time;
	
/*  Check parameters  */
	checkParamPointer(full_scale_adj, 2);
	
	checkErr(m228_ReadRegister16(handle, M228_CAL_EEPROM_CONTROL_REGISTER, &data));
	
	elasped_time = 0;
	start_time = (Real64)clock() / (Real64)CLOCKS_PER_SEC;
	while (((data & M228_I2C_OPC_BIT_MASK) == 0) && (elasped_time < 0.5))
	{
		checkErr(m228_ReadRegister16(handle, M228_CAL_EEPROM_CONTROL_REGISTER, &data));
		elasped_time = ((Real64)clock() / (Real64)CLOCKS_PER_SEC) - start_time;
	}
	if (elasped_time < 0.5)
	{	checkErr(m228_ReadRegister16(handle, M228_CAL_FULLSCALE_OFFSET_REGISTER, &data));
		*full_scale_adj = (data & M228_FSVAL_BIT_MASK) >> M228_FSVAL_BIT_NUM; 
	} 
	else
		error = M228_ERROR_I2C_TIMEOUT;
	
Error:
	return error;
}

 
/*****************************************************************************
*	m228_ReadFifo
*
*	Reads (if STM=0) the next timestamp/data pair from the fifo or (if STM=1) the 
* nxt data value from the fifo.  If no valid data	exists in the fifo this function 
* will return an error.
******************************************************************************/
Int32 m228_ReadFifo (M228_HANDLE handle, Bool width32bit, UInt32 *timestamp, Real64 *value, UInt16 *raw_data)
{	Int32 error = M228_SUCCESS;
	UInt16 data;
	
/*  Check parameters  */
	checkParamRange(width32bit, 0, 0, 2);	// only 16 bit currently supported

/*	Read the fifo to get the first word  */
	checkErr(m228_ReadRegister16(handle, M228_FIFO_DATA_REGISTER, &data));
	
/*  If data valid bit is set then read the fifo 2 more times  */	
	if (data & M228_DATA_VALID_BIT_MASK)  //Check valid bit
	{	if(raw_data != NULL)
			*raw_data = (data & M228_DATA_14_BIT_MASK);
		
		if(value != NULL)
			/* Calculate the real value  */		
			checkErr(m228_ConvertToVoltage(handle, &data, value, 1, 1, 0, 0, 0));

		if(timestamp != NULL) // if data only storage mode
		{	checkErr(m228_ReadRegister16(handle, M228_FIFO_DATA_REGISTER, &data));
			*timestamp = (data << 16) & 0xFFFF0000;
			checkErr(m228_ReadRegister16(handle, M228_FIFO_DATA_REGISTER, &data));
			*timestamp |= data;
		}
	}
	else
	{	error = M228_ERROR_FIFO_EMPTY;
		goto Error;
	}

Error:
	return error;
}


/*****************************************************************************
*	m228_ReadFifoBuffer
*
*	Reads a buffer of time stamp/data pairs from the fifo.  If no valid data
*	exists in the fifo this function will return an error.
******************************************************************************/
Int32 m228_ReadFifoBuffer (M228_HANDLE handle, Bool width32bit, UInt32 count, UInt32 *timestamp, Real64 *value, UInt16 *raw_data, UInt32 *num_read)
{	Int32 error = M228_SUCCESS;
	UInt16 reads_per_entry, *fifo_data=0, *lraw_data=0;
	UInt32 i;
	UInt32 local_num_read;
	Bool free_lraw_data = 0;

/*  Check parameters  */
	checkParamRange(width32bit, 0, 0, 2);	// only 16 bit currently supported

	if(timestamp == NULL)
		reads_per_entry = 1;	// value storage only mode
	else
		reads_per_entry = 3;
		
	fifo_data = malloc(count * sizeof(UInt16) * reads_per_entry);
	if(fifo_data == NULL)
	{	error = M228_ERROR_OUT_OF_MEMORY;
		goto Error;
	}
	
	if(raw_data == NULL)
	{	lraw_data = malloc(count * sizeof(UInt16));
		if(lraw_data == NULL)
		{	error = M228_ERROR_OUT_OF_MEMORY;
			goto Error;
		}
		free_lraw_data = 1;
	}
	
	local_num_read = 0;
	
/*	Read the fifo data port */
	apisCheckErr(apis_readfifo(handle->apis_handle, 2, M228_FIFO_DATA_REGISTER, count * reads_per_entry, fifo_data));
	
	for(i=0; i<count; i++)
	{
		if (fifo_data[i*reads_per_entry] & M228_DATA_VALID_BIT_MASK)  //Check valid bit
		{	if(raw_data == NULL)
				lraw_data[i] = (fifo_data[i*reads_per_entry] & M228_DATA_14_BIT_MASK);	// use local raw data array
			else
				raw_data[i] = (fifo_data[i*reads_per_entry] & M228_DATA_14_BIT_MASK);		// use external raw data array
			if(reads_per_entry == 3)	// this also means that timestamp != NULL
			{	timestamp[i] = (fifo_data[(i*reads_per_entry) + 1] << 16) & 0xFFFF0000;
				timestamp[i] |= fifo_data[(i*reads_per_entry) + 2];
			}
			local_num_read++;
		}
	}
	
	if(local_num_read != count)
		error = M228_WARNING_NUM_SAMPLES_NOT_AVAILABLE;

	if(num_read != NULL)
		*num_read = local_num_read;
	
	if((value != NULL) && (local_num_read > 0))
	{	if(raw_data == NULL)
			checkErr(m228_ConvertToVoltage(handle, lraw_data, value, local_num_read, 1, 0, 0, 0));
		else
			checkErr(m228_ConvertToVoltage(handle, raw_data, value, local_num_read, 1, 0, 0, 0));
	}
	
Error:
	if(fifo_data != NULL)
		free(fifo_data);
	if(free_lraw_data == 1)
		free(lraw_data);
	return error;
}


/*****************************************************************************
*	m228_ConvertToVoltage
*
******************************************************************************/
Int32 m228_ConvertToVoltage(M228_HANDLE handle, UInt16 *raw_data, Real64 *voltage, UInt32 num_points, Bool read_module, Bool divider, UInt8 fgain, UInt8 bgain)
{	Int32 error = M228_SUCCESS;
	Int16 raw_val_sign_ext;
	UInt32 i;

/*  Check parameters  */
	checkParamPointer(raw_data, 2);
	checkParamPointer(voltage, 3);
	checkParamRange(read_module, 0, 1, 5);
	checkParamRange(divider, 0, 1, 6);
	checkParamRange(fgain, 0, 3, 7);
	checkParamRange(bgain, 0, 7, 8);

	if(read_module)
	{	checkErr(m228_QueryAnalogFrontEndConfiguration (handle, &divider, &fgain, &bgain));
	}
	else
	{	checkParamRange(divider, 0, 1, 5);
		checkParamRange(fgain, 0, 3, 6);	
		checkParamRange(bgain, 0, 7, 7);	
	}
	
	for(i=0; i<num_points; i++)
	{	
		if(raw_data[i] == 0x2000)										 //These are just warning so don't exit
			error = M228_WARNING_UNDERVOLTAGE;	   
		if(raw_data[i] == 0x1FFF)
			error = M228_WARNING_OVERVOLTAGE;
		
		if(raw_data[i] & 0x2000)
			raw_val_sign_ext = raw_data[i] | 0xC000;
		else
			raw_val_sign_ext = raw_data[i];
	
		voltage[i] = (Real64)raw_val_sign_ext * (Real64)M228_A2D_RESOLUTION;
	
		if(bgain_multiplier[bgain] == 0)   //with a bgain of 0 the input is essentially disabled
			voltage[i] = 0;
		else	
		{	voltage[i] /= (Real64)bgain_multiplier[bgain];
			voltage[i] /= (Real64)fgain_multiplier[fgain];
	
			if(divider == 0)
				voltage[i] *= 10.0;
		}
	}
		
Error:
	return error;
}


/*****************************************************************************
*	m228_ConvertToRaw
*
******************************************************************************/
Int32 m228_ConvertToRaw(M228_HANDLE handle, Real64 *real_value , UInt16 *raw_data, UInt32 num_points, Bool read_module, Bool divider, UInt8 fgain, UInt8 bgain)
{	Int32 error = M228_SUCCESS;
	Int16 raw_val_sign_ext;
	Real64 voltage;
	UInt32 i;

/*  Check parameters  */
	checkParamPointer(real_value, 2);
	checkParamPointer(raw_data, 3);
	checkParamRange(read_module, 0, 1, 5);
	checkParamRange(divider, 0, 1, 6);
	checkParamRange(fgain, 0, 3, 7);
	checkParamRange(bgain, 0, 7, 8);
	
	if(read_module)
	{	checkErr(m228_QueryAnalogFrontEndConfiguration (handle, &divider, &fgain, &bgain));
	}
	else
	{	checkParamRange(divider, 0, 1, 5);
		checkParamRange(fgain, 0, 3, 6);	
		checkParamRange(bgain, 0, 7, 7);	
	}
	
	for(i=0; i<num_points; i++)
	{	
		if(divider == 0)
			voltage = real_value[i] / 10.0;
		else
			voltage = real_value[i];

		voltage *= (Real64)fgain_multiplier[fgain];
		voltage *= (Real64)bgain_multiplier[bgain];
	
		raw_val_sign_ext = (Int16)(voltage / (Real64)M228_A2D_RESOLUTION);

		if(raw_val_sign_ext & 0x8000)
			raw_data[i] = raw_val_sign_ext & ~0xC000;
		else
			raw_data[i] = raw_val_sign_ext;
	}	
Error:
	return error;
}


/*****************************************************************************
*	m228_ConfigureFunctionSource
*
******************************************************************************/
Int32 m228_ConfigureFunctionSource (M228_HANDLE handle, UInt8 function, UInt8 source, Bool invert)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamRange(function, 0, 3, 2);
	checkParamRange(source, 0, 7, 3);
	checkParamRange(invert, 0, 1, 4);

	checkErr(m228_ReadRegister16(handle, M228_FUNCTION_SOURCE_CONTROL_REGISTER, &data));

	switch (function)
	{
		case M228_FORCE_STORE_ALL_FUNCTION:
		{	data &= ~M228_FST_SOURCE_BIT_MASK;
			data |= (UInt16)(source << M228_FST_SOURCE_BIT_NUM);
			if(invert == M228_TRUE)
				data |= M228_FST_SOURCE_INVERT_BIT_MASK;
			else
				data &= ~M228_FST_SOURCE_INVERT_BIT_MASK;
		}
			break;
		case M228_APERTURE_SELECT_FUNCTION	:
		{	data &= ~M228_ASEL_SOURCE_BIT_MASK;
			data |= (UInt16)(source << M228_ASEL_SOURCE_BIT_NUM);
			if(invert == M228_TRUE)
				data |= M228_ASEL_SOURCE_INVERT_BIT_MASK;
			else
				data &= ~M228_ASEL_SOURCE_INVERT_BIT_MASK;
		}
			break;
		case M228_A2D_CONV_ENABLE_FUNCTION:
		{	data &= ~M228_ACE_SOURCE_BIT_MASK;
			data |= (UInt16)(source << M228_ACE_SOURCE_BIT_NUM);
			if(invert == M228_TRUE)
				data |= M228_ACE_SOURCE_INVERT_BIT_MASK;
			else
				data &= ~M228_ACE_SOURCE_INVERT_BIT_MASK;
		}
			break;
		case M228_TIMESTAMP_RUN_ENABLE_FUNCTION:
		{	data &= ~M228_TRUN_SOURCE_BIT_MASK;
			data |= (UInt16)(source << M228_TRUN_SOURCE_BIT_NUM);
			if(invert == M228_TRUE)
				data |= M228_TRUN_SOURCE_INVERT_BIT_MASK;
			else
				data &= ~M228_TRUN_SOURCE_INVERT_BIT_MASK;
		}
			break;
		default:
			break;
	}

	checkErr(m228_WriteRegister16(handle, M228_FUNCTION_SOURCE_CONTROL_REGISTER, data));

Error:
	return error;
}


/*****************************************************************************
*	m228_QueryFunctionSource
*
******************************************************************************/
Int32 m228_QueryFunctionSource (M228_HANDLE handle, UInt8 function, UInt8 *source, Bool *invert)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamRange(function, 0, 3, 2);
	checkParamPointer(source, 3);
	checkParamPointer(invert, 4);

	checkErr(m228_ReadRegister16(handle, M228_FUNCTION_SOURCE_CONTROL_REGISTER, &data));

	switch (function)
	{
		case M228_FORCE_STORE_ALL_FUNCTION:
			*source = (UInt8)((data & M228_FST_SOURCE_BIT_MASK) >> M228_FST_SOURCE_BIT_NUM);
			if(data & M228_FST_SOURCE_INVERT_BIT_MASK)
				*invert = M228_TRUE;
			else
				*invert = M228_FALSE;
			break;
		case M228_APERTURE_SELECT_FUNCTION	:
			*source = (UInt8)((data & M228_ASEL_SOURCE_BIT_MASK) >> M228_ASEL_SOURCE_BIT_NUM);
			if(data & M228_ASEL_SOURCE_INVERT_BIT_MASK)
				*invert = M228_TRUE;
			else
				*invert = M228_FALSE;
			break;
		case M228_A2D_CONV_ENABLE_FUNCTION:
			*source = (UInt8)((data & M228_ACE_SOURCE_BIT_MASK) >> M228_ACE_SOURCE_BIT_NUM);
			if(data & M228_ACE_SOURCE_INVERT_BIT_MASK)
				*invert = M228_TRUE;
			else
				*invert = M228_FALSE;
			break;
		case M228_TIMESTAMP_RUN_ENABLE_FUNCTION:
			*source = (UInt8)((data & M228_TRUN_SOURCE_BIT_MASK) >> M228_TRUN_SOURCE_BIT_NUM);
			if(data & M228_TRUN_SOURCE_INVERT_BIT_MASK)
				*invert = M228_TRUE;
			else
				*invert = M228_FALSE;
			break;
		default:
			break;
	}


Error:
	return error;
}


/*****************************************************************************
*	m228_ConfigureFrontPanelOutput
*
******************************************************************************/
Int32 m228_ConfigureFrontPanelOutput (M228_HANDLE handle, Bool channel, UInt8 source, Bool invert)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamRange(channel, 0, 1, 2);
	checkParamRange(source, 0, 7, 3);
	checkParamRange(invert, 0, 1, 4);

	checkErr(m228_ReadRegister16(handle, M228_OUTPUT_SOURCE_CONTROL_REGISTER, &data));

	if(channel == M228_OUTPUT_A)
	{	
		data &= ~M228_FPOUTA_SOURCE_BIT_MASK;
		data |= (UInt16)(source << M228_FPOUTA_SOURCE_BIT_NUM);
		if(invert == M228_TRUE)
			data |= M228_FPOUTA_INVERT_BIT_MASK;
		else
			data &= ~M228_FPOUTA_INVERT_BIT_MASK;
	}
	else
	{	data &= ~M228_FPOUTB_SOURCE_BIT_MASK;
		data |= (UInt16)(source << M228_FPOUTB_SOURCE_BIT_NUM);
		if(invert == M228_TRUE)
			data |= M228_FPOUTB_INVERT_BIT_MASK;
		else
			data &= ~M228_FPOUTB_INVERT_BIT_MASK;
	}

	checkErr(m228_WriteRegister16(handle, M228_OUTPUT_SOURCE_CONTROL_REGISTER, data));

Error:
	return error;
}


/*****************************************************************************
*	m228_QueryFrontPanelOutput
*
******************************************************************************/
Int32 m228_QueryFrontPanelOutput (M228_HANDLE handle, Bool channel, UInt8 *source, Bool *invert)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamRange(channel, 0, 1, 2);
	checkParamPointer(source, 3);
	checkParamPointer(invert, 4);

	checkErr(m228_ReadRegister16(handle, M228_OUTPUT_SOURCE_CONTROL_REGISTER, &data));

	if(channel == M228_OUTPUT_A)
	{	*source = (UInt8)((data & M228_FPOUTA_SOURCE_BIT_MASK) >> M228_FPOUTA_SOURCE_BIT_NUM);
		if(data & M228_FPOUTA_INVERT_BIT_MASK)
			*invert = M228_TRUE;
		else
			*invert = M228_FALSE;
	}
	else
	{	*source = (UInt8)((data & M228_FPOUTB_SOURCE_BIT_MASK) >> M228_FPOUTB_SOURCE_BIT_NUM);
		if(data & M228_FPOUTB_INVERT_BIT_MASK)
			*invert = M228_TRUE;
		else
			*invert = M228_FALSE;
	}

Error:
	return error;
}


/*****************************************************************************
*	m228_ConfigureBackplaneTriggerOutput
*
******************************************************************************/
Int32 m228_ConfigureBackplaneTriggerOutput (M228_HANDLE handle, Bool channel, UInt8 source, Bool invert)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamRange(channel, 0, 1, 2);
	checkParamRange(source, 0, 7, 3);
	checkParamRange(invert, 0, 1, 4);

	checkErr(m228_ReadRegister16(handle, M228_OUTPUT_SOURCE_CONTROL_REGISTER, &data));

	if(channel == M228_OUTPUT_A)
	{	data &= ~M228_TRIG_OUTA_SOURCE_BIT_MASK;
		data |= (UInt16)(source << M228_TRIG_OUTA_SOURCE_BIT_NUM);
		if(invert == M228_TRUE)
			data |= M228_TRIG_OUTA_INVERT_BIT_MASK;
		else
			data &= ~M228_TRIG_OUTA_INVERT_BIT_MASK;
	}
	else
	{	data &= ~M228_TRIG_OUTB_SOURCE_BIT_MASK;
		data |= (UInt16)(source << M228_TRIG_OUTB_SOURCE_BIT_NUM);
		if(invert == M228_TRUE)
			data |= M228_TRIG_OUTB_INVERT_BIT_MASK;
		else
			data &= ~M228_TRIG_OUTB_INVERT_BIT_MASK;
	}

	checkErr(m228_WriteRegister16(handle, M228_OUTPUT_SOURCE_CONTROL_REGISTER, data));

Error:
	return error;
}


/*****************************************************************************
*	m228_QueryBackplaneTriggerOutput
*
******************************************************************************/
Int32 m228_QueryBackplaneTriggerOutput (M228_HANDLE handle, Bool channel, UInt8 *source, Bool *invert)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamRange(channel, 0, 1, 2);
	checkParamPointer(source, 3);
	checkParamPointer(invert, 4);

	checkErr(m228_ReadRegister16(handle, M228_OUTPUT_SOURCE_CONTROL_REGISTER, &data));

	if(channel == M228_OUTPUT_A)
	{	*source = (UInt8)((data & M228_TRIG_OUTA_SOURCE_BIT_MASK) >> M228_TRIG_OUTA_SOURCE_BIT_NUM);
		if(data & M228_TRIG_OUTA_INVERT_BIT_MASK)
			*invert = M228_TRUE;
		else
			*invert = M228_FALSE;
	}
	else
	{	*source = (UInt8)((data & M228_TRIG_OUTB_SOURCE_BIT_MASK) >> M228_TRIG_OUTB_SOURCE_BIT_NUM);
		if(data & M228_TRIG_OUTB_INVERT_BIT_MASK)
			*invert = M228_TRUE;
		else
			*invert = M228_FALSE;
	}

Error:
	return error;
}


/*****************************************************************************
*	m228_ConfigureFrontPanelInput
*
******************************************************************************/
Int32 m228_ConfigureFrontPanelInput (M228_HANDLE handle, Bool channel, Bool impedance, Real64 threshold, Bool invert, Bool latch)
{	Int32 error = M228_SUCCESS;
	UInt16 data;
 	UInt16 vbits;

/*  Check parameters  */
	checkParamRange(channel, 0, 1, 2);
	checkParamRange(impedance, 0, 1, 3);
	checkParamRange(threshold, 0.0, 5.0, 4);
	checkParamRange(invert, 0, 1, 5);
	checkParamRange(latch, 0, 1, 6);

	if(channel == M228_INPUT_A)
		checkErr(m228_ReadRegister16(handle, M228_INPUT_A_CONTROL_REGISTER, &data));
	else											
		checkErr(m228_ReadRegister16(handle, M228_INPUT_B_CONTROL_REGISTER, &data));

	if(impedance == M228_TRUE)
		data |= M228_FPIN_IMPEDANCE_BIT_MASK;
	else
		data &= ~M228_FPIN_IMPEDANCE_BIT_MASK;

	if(invert == M228_TRUE)
		data |= M228_FPIN_INVERT_BIT_MASK;
	else
		data &= ~M228_FPIN_INVERT_BIT_MASK;

	if(latch == M228_TRUE)
		data |= M228_FPIN_LATCH_BIT_MASK;
	else
		data &= ~M228_FPIN_LATCH_BIT_MASK;

	vbits = (UInt16)(((threshold + M228_FPIN_THRESHOLD_OFFSET) / M228_FPIN_THRESHOLD_RESOLUTION) + 0.5);		 //add 0.5 to round to nearest int
	data &= ~M228_FPIN_THRESHOLD_BIT_MASK;
	data |= M228_FPIN_THRESHOLD_BIT_MASK & vbits;

	if(channel == M228_INPUT_A)
		checkErr(m228_WriteRegister16(handle, M228_INPUT_A_CONTROL_REGISTER, data & ~M228_CLEAR_LATCHED_BIT_MASK));
	else
		checkErr(m228_WriteRegister16(handle, M228_INPUT_B_CONTROL_REGISTER, data & ~M228_CLEAR_LATCHED_BIT_MASK));

Error:
	return error;
}


/*****************************************************************************
*	m228_QueryFrontPanelInput
*
******************************************************************************/
Int32 m228_QueryFrontPanelInput (M228_HANDLE handle, Bool channel, Bool *impedance, Real64 *threshold, Bool *invert, Bool *latch)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamRange(channel, 0, 1, 2);
	checkParamPointer(impedance, 3);
	checkParamPointer(threshold, 4);
	checkParamPointer(invert, 5);
	checkParamPointer(latch, 6);

	if(channel == M228_INPUT_A)
		checkErr(m228_ReadRegister16(handle, M228_INPUT_A_CONTROL_REGISTER, &data));
	else											
		checkErr(m228_ReadRegister16(handle, M228_INPUT_B_CONTROL_REGISTER, &data));

	if(data & M228_FPIN_IMPEDANCE_BIT_MASK)
		*impedance = M228_TRUE;
	else
		*impedance = M228_FALSE;

	if(data & M228_FPIN_INVERT_BIT_MASK)
		*invert = M228_TRUE;
	else
		*invert = M228_FALSE;

	if(data & M228_FPIN_LATCH_BIT_MASK)
		*latch = M228_TRUE;
	else
		*latch = M228_FALSE;

	*threshold = ((Real64)(data & M228_FPIN_THRESHOLD_BIT_MASK) * M228_FPIN_THRESHOLD_RESOLUTION) - M228_FPIN_THRESHOLD_OFFSET;

Error:
	return error;
}


/*****************************************************************************
*	m228_ConfigureBackplaneTriggerInput
*
******************************************************************************/
Int32 m228_ConfigureBackplaneTriggerInput (M228_HANDLE handle, Bool channel, Bool invert, Bool latch)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamRange(channel, 0, 1, 2);
	checkParamRange(invert, 0, 1, 3);
	checkParamRange(latch, 0, 1, 4);

	if(channel == M228_INPUT_A)
		checkErr(m228_ReadRegister16(handle, M228_INPUT_A_CONTROL_REGISTER, &data));
	else											
		checkErr(m228_ReadRegister16(handle, M228_INPUT_B_CONTROL_REGISTER, &data));

	if(invert == M228_TRUE)
		data |= M228_TRIG_INVERT_BIT_MASK;
	else
		data &= ~M228_TRIG_INVERT_BIT_MASK;

	if(latch == M228_TRUE)
		data |= M228_TRIG_LATCH_BIT_MASK;
	else
		data &= ~M228_TRIG_LATCH_BIT_MASK;

	if(channel == M228_INPUT_A)
		checkErr(m228_WriteRegister16(handle, M228_INPUT_A_CONTROL_REGISTER, data & ~M228_CLEAR_LATCHED_BIT_MASK));
	else
		checkErr(m228_WriteRegister16(handle, M228_INPUT_B_CONTROL_REGISTER, data & ~M228_CLEAR_LATCHED_BIT_MASK));

Error:
	return error;
}


/*****************************************************************************
*	m228_QueryBackplaneTriggerInput
*
******************************************************************************/
Int32 m228_QueryBackplaneTriggerInput (M228_HANDLE handle, Bool channel, Bool *invert, Bool *latch)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamRange(channel, 0, 1, 2);
	checkParamPointer(invert, 3);
	checkParamPointer(latch, 4);

	if(channel == M228_INPUT_A)
		checkErr(m228_ReadRegister16(handle, M228_INPUT_A_CONTROL_REGISTER, &data));
	else											
		checkErr(m228_ReadRegister16(handle, M228_INPUT_B_CONTROL_REGISTER, &data));

	if(data & M228_TRIG_INVERT_BIT_MASK)
		*invert = M228_TRUE;
	else
		*invert = M228_FALSE;

	if(data & M228_TRIG_LATCH_BIT_MASK)
		*latch = M228_TRUE;
	else
		*latch = M228_FALSE;

Error:
	return error;
}


/*****************************************************************************
*	m228_QueryInputState
*
******************************************************************************/
Int32 m228_QueryInputState (M228_HANDLE handle, Bool *fpina, Bool *fpinb, Bool *triga, Bool *trigb)
{	Int32 error = M228_SUCCESS;
	UInt16 data;
	
	if((fpina != NULL) || (triga != NULL))
	{	checkErr(m228_ReadRegister16(handle, M228_INPUT_A_CONTROL_REGISTER, &data));
	
		if(fpina != NULL)
		{	if(data & M228_FPIN_STATUS_BIT_MASK)
				*fpina = M228_TRUE;
			else
				*fpina = M228_FALSE;
		}
		
		if(triga != NULL)
		{	if(data & M228_TRIG_STATUS_BIT_MASK)
				*triga = M228_TRUE;
			else
				*triga = M228_FALSE;
		}
	}
	
	if((fpinb != NULL) || (trigb != NULL))
	{	checkErr(m228_ReadRegister16(handle, M228_INPUT_B_CONTROL_REGISTER, &data));

		if(fpinb != NULL)
		{	if(data & M228_FPIN_STATUS_BIT_MASK)
				*fpinb = M228_TRUE;
			else
				*fpinb = M228_FALSE;
		}

		if(trigb != NULL)
		{	if(data & M228_TRIG_STATUS_BIT_MASK)
				*trigb = M228_TRUE;
			else
				*trigb = M228_FALSE;
		}
	}
	
Error:
	return error;
}


/*****************************************************************************
*	m228_ClearLatchedInput
*
******************************************************************************/
Int32 m228_ClearLatchedInput (M228_HANDLE handle, UInt8 input)
{	Int32 error = M228_SUCCESS;
	UInt16 data;					
	
/*  Check parameters  */
	checkParamRange(input, M228_FPINA_SIGNAL, M228_MTRIGINB_SIGNAL, 2);
	
	if ((input == M228_FPINA_SIGNAL) || (input == M228_MTRIGINA_SIGNAL))
	{	checkErr(m228_ReadRegister16(handle, M228_INPUT_A_CONTROL_REGISTER, &data));
		if (input == M228_MTRIGINA_SIGNAL)
			data |= M228_CLEAR_LATCHED_TRIG_BIT_MASK;
		else
			data |= M228_CLEAR_LATCHED_FPIN_BIT_MASK;
		checkErr(m228_WriteRegister16(handle, M228_INPUT_A_CONTROL_REGISTER, data));
	}
	else
	{	checkErr(m228_ReadRegister16(handle, M228_INPUT_B_CONTROL_REGISTER, &data));
		if (input == M228_MTRIGINB_SIGNAL)
			data |= M228_CLEAR_LATCHED_TRIG_BIT_MASK;
		else
			data |= M228_CLEAR_LATCHED_FPIN_BIT_MASK;
		checkErr(m228_WriteRegister16(handle, M228_INPUT_B_CONTROL_REGISTER, data));
	}

Error:
	return error;
}


/*****************************************************************************
*	m228_ConfigureA2DTemperature
*
******************************************************************************/
Int32 m228_ConfigureA2DTemperature (M228_HANDLE handle, Bool enable)
{	Int32 error = M228_SUCCESS;

/*  Check parameters  */
	checkParamRange(enable, 0, 1, 2);
	
	if (enable == M228_TRUE)
		checkErr(m228_WriteRegister16(handle, M228_A2D_TEMPERATURE_REGISTER, 0));
	else
		checkErr(m228_WriteRegister16(handle, M228_A2D_TEMPERATURE_REGISTER, M228_DISABLE_TEMP_READ_BIT_MASK));

Error:
	return error;
}


/*****************************************************************************
*	m228_QueryA2DTemperature
*
******************************************************************************/
Int32 m228_QueryA2DTemperature (M228_HANDLE handle, Real64 *temperature, Bool *enabled)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  Check parameters  */
	checkParamPointer(temperature, 2);
	checkParamPointer(enabled, 3);
	
	checkErr(m228_ReadRegister16(handle, M228_A2D_TEMPERATURE_REGISTER, &data));
	if ((data & 0x0200) == 0)	// positive temperature
		*temperature = (Real64)(data & 0x1FF) / 4.0;
	else											// negative temperature
		*temperature = ((Real64)(data & 0x1FF) - 512.0) / 4.0;
	if ((data & M228_DISABLE_TEMP_READ_BIT_MASK) == 0)
		*enabled = 1;
	else
		*enabled = 0;

Error:
	return error;
}


/*****************************************************************************
*	m228_WriteEEPROMValue
*
******************************************************************************/
Int32 m228_WriteEEPROMValue (M228_HANDLE handle, UInt16 address, UInt16 data)
{	Int32 error = M228_SUCCESS;
	UInt16 opc_data;
	Real64 start_time, elasped_time;
	Int32 count;
	
/*  Check parameters  */
	checkParamRange(address, 0, M228_EEPROM_ADDR_BIT_MASK, 2);
		
	checkErr(m228_WriteRegister16(handle, M228_CAL_EEPROM_DATA_REGISTER, data));
	checkErr(m228_WriteRegister16(handle, M228_CAL_EEPROM_CONTROL_REGISTER, M228_EEPROM_ADDR_BIT_MASK & address));
	checkErr(m228_WriteRegister16(handle, M228_CAL_EEPROM_CONTROL_REGISTER, M228_WRITE_EEPROM_BIT_MASK | address));
	
	count = 0;
	elasped_time = 0.0;
	start_time = (Real64)clock() / (Real64)CLOCKS_PER_SEC;
	checkErr(m228_ReadRegister16(handle, M228_CAL_EEPROM_CONTROL_REGISTER, &opc_data));
	while (((opc_data & M228_I2C_OPC_BIT_MASK) == 0) && (elasped_time < 0.5))
	{ checkErr(m228_ReadRegister16(handle, M228_CAL_EEPROM_CONTROL_REGISTER, &opc_data));
		elasped_time = ((Real64)clock() / (Real64)CLOCKS_PER_SEC) - start_time;
		count++;
	}
	if (elasped_time > 0.5)
		error = M228_ERROR_I2C_TIMEOUT;

/*  The Write can take a maximum of 5ms to complete following the OPC  */	
	elasped_time = 0.0;
	start_time = (Real64)clock() / (Real64)CLOCKS_PER_SEC;
	while (elasped_time < 0.005)
	{	elasped_time = ((Real64)clock() / (Real64)CLOCKS_PER_SEC) - start_time;
	}														
	
	
Error:
	return error;
}


/*****************************************************************************
*	m228_ReadEEPROMValue
*
******************************************************************************/
Int32 m228_ReadEEPROMValue (M228_HANDLE handle, UInt16 address, UInt16 *data)
{	Int32 error = M228_SUCCESS;
	UInt16 opc_data;
	Real64 start_time, elasped_time;
	
/*  Check parameters  */
	checkParamRange(address, 0, M228_EEPROM_ADDR_BIT_MASK, 2);
	checkParamPointer(data, 3);
	
	checkErr(m228_WriteRegister16(handle, M228_CAL_EEPROM_CONTROL_REGISTER, M228_EEPROM_ADDR_BIT_MASK & address));
	checkErr(m228_WriteRegister16(handle, M228_CAL_EEPROM_CONTROL_REGISTER, M228_READ_EEPROM_BIT_MASK | address));

	elasped_time = 0;
	start_time = (Real64)clock() / (Real64)CLOCKS_PER_SEC;
	checkErr(m228_ReadRegister16(handle, M228_CAL_EEPROM_CONTROL_REGISTER, &opc_data));
	while (((opc_data & M228_I2C_OPC_BIT_MASK) == 0) && (elasped_time < 0.5))
	{ checkErr(m228_ReadRegister16(handle, M228_CAL_EEPROM_CONTROL_REGISTER, &opc_data));
		elasped_time = ((Real64)clock() / (Real64)CLOCKS_PER_SEC) - start_time;
	}
	if (elasped_time < 0.5)
		checkErr(m228_ReadRegister16(handle, M228_CAL_EEPROM_DATA_REGISTER, data));
	else
		error = M228_ERROR_I2C_TIMEOUT;
	
Error:
	return error;
}


/*****************************************************************************
*	chm228_FormatEEPROM
*
******************************************************************************/
Int32 m228_FormatEEPROM (M228_HANDLE handle)
{	Int32 error = M228_SUCCESS;
	const UInt8 instrument_name[16] = M228_CALIBRATION_EEPROM_HEADER_STRING;
	Int32 i;
	
/*  Clear the EEPROM Header  */
	for(i=0; i<15; i+=2)	
	{	checkErr(m228_WriteEEPROMValue(handle, i, 0xFFFF));				
	}

/*  Write EEPROM Header Information  */	
	for(i=0; i<15; i+=2)	
	{	checkErr(m228_WriteEEPROMValue(handle, i, *(UInt16 *)(&instrument_name[i])));				//this also writes the num_entries field to 0 when i=14
	}
	
/*  Invalidate all Calibration Entries by making first byte of entry label 0xFF  */
	for(i=0; i<15; i++)
	{	checkErr(m228_WriteEEPROMValue(handle, 0x10 + (0x20 * i), 0xFFFF));
	}

Error:
	return error;
}


/*****************************************************************************
*	m228_QueryEEPROMFormatted
*
******************************************************************************/
Int32 m228_QueryEEPROMFormatted (M228_HANDLE handle, Bool *formatted)
{	Int32 error = M228_SUCCESS;
	Int8 instrument_name[16]="";
	Int32 i;

/*  Check parameters  */
	checkParamPointer(formatted, 2);
	
	for(i=0; i<15; i+=2)	
	{	checkErr(m228_ReadEEPROMValue(handle, i, (UInt16 *)(&instrument_name[i])));
	}
	instrument_name[15] = '\0';  //must terminate the string
	
	if(strcmp(instrument_name, M228_CALIBRATION_EEPROM_HEADER_STRING) == 0)
		*formatted = M228_TRUE;
	else
		*formatted = M228_FALSE;
	
Error:
	return error;
}


/*****************************************************************************
*	m228_GetCalibrationEntry
*
******************************************************************************/
Int32 m228_GetCalibrationEntry (M228_HANDLE handle, UInt8 entry_num, M228_CALIBRATION_ENTRY *entry)
{	Int32 error = M228_SUCCESS;
	Int32 i;
	UInt8 data_buff[32] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	UInt16 entry_offset;
	
/*  Check parameters  */
	checkParamRange(entry_num, 1, 15, 2);
	checkParamPointer(entry, 3);

	entry_offset = 0x10 + (0x20 * (entry_num - 1));
	for(i=0; i<32; i+=2)	
	{	checkErr(m228_ReadEEPROMValue(handle, entry_offset + i, (UInt16 *)(&data_buff[i])));
	}

	if(data_buff[0] == 0xFF)
		error = M228_WARNING_CALIBRATION_ENTRY_EMPTY;
	
	
	for(i=0; i<11; i++)
		entry->label[i] = data_buff[i];
	entry->label[11] = '\0';

	for(i=11; i<19; i++)	
		entry->date[i-11] = data_buff[i];
	entry->date[8] = '\0';;
	
	for(i=19; i<25; i++)
		entry->time[i-19] = data_buff[i];
	entry->time[6] = '\0';
	
	entry->temperature = data_buff[25]; 
	entry->cal_type = data_buff[26];
	entry->front_end_config = data_buff[27];
	entry->filter_config = data_buff[28];
	entry->offset_cal_value = data_buff[29];
	entry->full_scale_cal_value = data_buff[30];
	
Error:
	return error;
}


/*****************************************************************************
*	m228_StoreCalibrationEntry
*
******************************************************************************/
Int32 m228_StoreCalibrationEntry (M228_HANDLE handle, UInt8 entry_num, M228_CALIBRATION_ENTRY *entry)
{	Int32 error = M228_SUCCESS;
	Int32 i;
	UInt8 data_buff[32];
	UInt16 entry_offset;
	
/*  Check parameters  */
	checkParamRange(entry_num, 1, 15, 2);
	checkParamPointer(entry, 3);

	for(i=0; i<11; i++)
		data_buff[i] = entry->label[i];
	for(i=11; i<19; i++)	
		data_buff[i] = entry->date[i-11];
	for(i=19; i<25; i++)
		data_buff[i] = entry->time[i-19];
	data_buff[25] = entry->temperature;
	data_buff[26] = entry->cal_type;
	data_buff[27] = entry->front_end_config;
	data_buff[28] = entry->filter_config;
	data_buff[29] = entry->offset_cal_value;
	data_buff[30] = entry->full_scale_cal_value;
	data_buff[31] = 0;

	entry_offset = 0x10 + (0x20 * (entry_num - 1));
	for(i=0; i<32; i+=2)
	{	checkErr(m228_WriteEEPROMValue(handle, entry_offset + i, *(UInt16 *)(&data_buff[i])));
	}

Error:
	return error;
}


/*****************************************************************************
*	m228_RevisionQuery
*
*	Returns the version number of this driver.  
******************************************************************************/
Int32 m228_RevisionQuery (Int8 driver_rev[])
{	Int32 error = M228_SUCCESS;

/*  Check parameters  */
	checkParamPointer(driver_rev, 2);
	
	sprintf(driver_rev, "%s", M228_DRIVER_REV);
Error:
	return error;
}



/*****************************************************************************
*	m228_ErrorMessage
*
*	Returns a text description of the specified status code.  
******************************************************************************/
Int32 m228_ErrorMessage (Int32 statusCode, Int8 errorMessage[])
{	Int32 error = M228_SUCCESS;
	
/*  Check parameters  */
	checkParamPointer(errorMessage, 3);

/*  Check driver defined error codes  */
	if((statusCode >= M228_SPECIFIC_ERROR_BASE) && (statusCode < M228_SPECIFIC_ERROR_END))
	{	strcpy(errorMessage, errorMessageTable[statusCode - M228_SPECIFIC_ERROR_BASE]);
	}
	else if((statusCode >= M228_SPECIFIC_WARNING_BASE) && (statusCode < M228_SPECIFIC_WARNING_END))
	{	strcpy(errorMessage, warningMessageTable[statusCode - M228_SPECIFIC_WARNING_BASE]);
	}
	else if (statusCode == 0)
		sprintf(errorMessage, "No Error Occured");
	else
		sprintf(errorMessage, "Unknown Error Code [0x%x]", statusCode);
		
Error:
	return error;
}


/*****************************************************************************
*	m228_GetIdentWord
*
*	Reads the specified offset from the module's ID space and returns
*	the result.
******************************************************************************/
Int32 m228_GetIdentWord (M228_HANDLE handle, UInt16 offset, UInt16 *value)
{	Int32 error = M228_SUCCESS;
	UInt8 data;

/*  Check parameters  */
	checkParamPointer(value, 3);

	*value = 0;
	
/*  Send the read command to the serial EEPROM  */	
	checkErr(m228_IDWriteByte(handle, 0x80 | offset));
	
/*  Read upper byte  */	
	checkErr(m228_IDReadByte(handle, &data));
	*value = data << 8;
	
/*  Read lower byte  */	
	checkErr(m228_IDReadByte(handle, &data));
	*value |= data;
	
Error:	
	m228_WriteRegister16(handle, M228_ID_ACCESS_REGISTER, 0);
	return error;
}


/*****************************************************************************
*	m228_EnableInterrupts 
*
*	Configures the m228 to produce interrupts.  
******************************************************************************/
Int32 m228_EnableInterrupts (M228_HANDLE handle,  Bool fifoFull,  Bool fifoHalfFull, Bool fifoQtrFull, Bool fifoNotEmpty, Bool timestampRollover)
{	Int32 error = M228_SUCCESS;
	UInt16 value = 0;
	
/*  Check parameters  */
	checkParamRange(fifoFull, 0, 1, 2);
	checkParamRange(fifoHalfFull, 0, 1, 3);
	checkParamRange(fifoQtrFull, 0, 1, 4);
	checkParamRange(fifoNotEmpty, 0, 1, 5);
	checkParamRange(timestampRollover, 0, 1, 6);


/*  Configure the Ma203  */	
	checkErr(m228_ReadRegister16(handle, M228_INTERRUPT_CONTROL_REGISTER, &value));
	value &= ~M228_INTERRUPT_ENABLE_BIT_MASK;
	value |= (fifoFull << 7) | (fifoHalfFull << 6) | (fifoQtrFull << 5) | (fifoNotEmpty << 4) | (timestampRollover << 3) | M228_INTERRUPT_GLOBAL_ENABLE_BIT_MASK;
	checkErr(m228_WriteRegister16(handle, M228_INTERRUPT_CONTROL_REGISTER, value));

Error:
	return error;
}


/*****************************************************************************
*	m228_QueryInterruptsEnable 
*
******************************************************************************/
Int32 m228_QueryInterruptsEnable (M228_HANDLE handle,  Bool *fifoFull,  Bool *fifoHalfFull, Bool *fifoQtrFull, Bool *fifoNotEmpty, Bool *timestampRollover)
{	Int32 error = M228_SUCCESS;
	UInt16 value = 0;
	
/*  Check parameters  */
	checkParamPointer(fifoFull, 2);
	checkParamPointer(fifoHalfFull, 3);
	checkParamPointer(fifoQtrFull, 4);
	checkParamPointer(fifoNotEmpty, 5);
	checkParamPointer(timestampRollover, 6);


/*  Configure the Ma203  */	
	checkErr(m228_ReadRegister16(handle, M228_INTERRUPT_CONTROL_REGISTER, &value));
	
	if(value & M228_INTERRUPT_FIFO_FULL_ENABLE_MASK)
		*fifoFull = M228_TRUE;
	else
		*fifoFull = M228_FALSE;
	

	if(value & M228_INTERRUPT_FIFO_HALF_FULL_ENABLE_MASK)
		*fifoHalfFull = M228_TRUE;
	else
		*fifoHalfFull = M228_FALSE;

	if(value & M228_INTERRUPT_FIFO_QTR_FULL_ENABLE_MASK)
		*fifoQtrFull = M228_TRUE;
	else
		*fifoQtrFull = M228_FALSE;

	if(value & M228_INTERRUPT_FIFO_NOT_EMPTY_ENABLE_MASK)
		*fifoNotEmpty = M228_TRUE;
	else
		*fifoNotEmpty = M228_FALSE;

	if(value & M228_INTERRUPT_TIMESTAMP_ROLLOVER_ENABLE_MASK)
		*timestampRollover = M228_TRUE;
	else
		*timestampRollover = M228_FALSE;

Error:
	return error;
}


/*****************************************************************************
*	m228_ClearInterrupts
*
*	Clears interrupt pendings for the channels specified by the
*	clearInterruptMask parameter.
******************************************************************************/
Int32 m228_ClearInterrupts (M228_HANDLE handle,  Bool fifoFull,  Bool fifoHalfFull, Bool fifoQtrFull, Bool fifoNotEmpty, Bool timestampRollover)
{	Int32 error = M228_SUCCESS;
	UInt16 value;

/*  Check parameters  */
	checkParamRange(fifoFull, 0, 1, 2);
	checkParamRange(fifoHalfFull, 0, 1, 3);
	checkParamRange(fifoQtrFull, 0, 1, 4);
	checkParamRange(fifoNotEmpty, 0, 1, 5);
	checkParamRange(timestampRollover, 0, 1, 6);

/*  Read current value of the control/status register  */
	checkErr(m228_ReadRegister16(handle, M228_INTERRUPT_CONTROL_REGISTER, &value));
	
/*  Set bits to clear interrupts  */	
	value &= ~M228_INTERRUPT_PENDING_BIT_MASK;
	value |= (fifoFull << M228_INTERRUPT_FIFO_FULL_BIT_NUM) | (fifoHalfFull << M228_INTERRUPT_FIFO_HALF_FULL_BIT_NUM) | (fifoQtrFull << M228_INTERRUPT_FIFO_QTR_FULL_BIT_NUM) | (fifoNotEmpty << M228_INTERRUPT_FIFO_NOT_EMPTY_BIT_NUM) | (timestampRollover << M228_INTERRUPT_TIMESTAMP_ROLLOVER_BIT_NUM); 
	
/*  Write new value to the register  */	
	checkErr(m228_WriteRegister16(handle, M228_INTERRUPT_CONTROL_REGISTER, value));

Error:
	return error;
}


/*****************************************************************************
*	m228_GetInterruptPending
*
*	Returns the value of the interrupt pending register
******************************************************************************/
Int32 m228_GetInterruptPending (M228_HANDLE handle,  Bool *fifoFull,  Bool *fifoHalfFull, Bool *fifoQtrFull, Bool *fifoNotEmpty, Bool *timestampRollover)
{	Int32 error = M228_SUCCESS;
	UInt16 value;
	
/*  Check parameters  */
	checkParamPointer(fifoFull, 2);
	checkParamPointer(fifoHalfFull, 3);
	checkParamPointer(fifoQtrFull, 4);
	checkParamPointer(fifoNotEmpty, 5);
	checkParamPointer(timestampRollover, 6);

	
	*fifoFull = 0;
	*fifoHalfFull = 0;
	*fifoQtrFull = 0;
	*fifoNotEmpty = 0;
	*timestampRollover = 0;
	
/*  Read control status register  */
	checkErr(m228_ReadRegister16(handle, M228_INTERRUPT_CONTROL_REGISTER, &value));
	
/*  Check flag bits  */	
	if(value & M228_FIFO_FULL_BIT_MASK)
		*fifoFull = 1;
	if(value & M228_FIFO_HALF_FULL_BIT_MASK)
		*fifoHalfFull = 1;
	if(value & M228_FIFO_QTR_FULL_BIT_MASK)
		*fifoQtrFull = 1;
	if(value & M228_FIFO_NOT_EMPTY_BIT_MASK)
		*fifoNotEmpty = 1;
	if(value & M228_TSR_BIT_MASK)
		*timestampRollover= 1;
Error:
	return error;
}


/*****************************************************************************
*	m228_IDWriteByte
*
*	Write a command byte to the ID PROM of the instrument.
*	The instruments ID PROM is serially accessed
*****************************************************************************/
static Int32 m228_IDWriteByte(M228_HANDLE handle, UInt8 value)
{	Int32 error = M228_SUCCESS;
	Int32 i;
	
/*  Initialize  */	
	checkErr(m228_WriteRegister16(handle, M228_ID_ACCESS_REGISTER, 0));
	checkErr(m228_WriteRegister16(handle, M228_ID_ACCESS_REGISTER, 0x4));
	
/*  Write start bit  */	
	checkErr(m228_IDWriteBit(handle, 1));
	for(i=0; i<=7; i++)
	{
	/*  Write data bits  */		
		checkErr(m228_IDWriteBit(handle, ((value &0x80)>>7)));
		value = value << 1;
	}
Error:	
	return error;
}


/*****************************************************************************
*	m228_IDWriteBit
*
*	Write a single bit to the ID PROM of the instrument.
*	The instruments ID PROM is serially accessed
*****************************************************************************/
static Int32 m228_IDWriteBit(M228_HANDLE handle,  Bool value)
{	Int32 error = M228_SUCCESS;

/*  data with clock low  */
	checkErr(m228_WriteRegister16(handle, M228_ID_ACCESS_REGISTER, (0x4 | (value & 0x1))));	

/*  data with clock high  */	 
	checkErr(m228_WriteRegister16(handle, M228_ID_ACCESS_REGISTER, (0x6 | (value & 0x1))));	
Error:	
	return error;
}


/*****************************************************************************
*	m228_IDReadByte
*
*	Reads a byte from the ID PROM of the instrument.
*	The instruments ID PROM is serially accessed
*****************************************************************************/
static Int32 m228_IDReadByte(M228_HANDLE handle, UInt8 *value)
{	Int32 error = M228_SUCCESS;
	Int32 i;
	 Bool data;
	
	*value = 0;
	for(i=7; i>=0; i--)
	{ 
	/*  read data 1 bit at a time  */	
		checkErr(m228_IDReadBit(handle, &data));
		*value |= ((data & 0x1) << i);
	}
	
	
Error:	
	return error;
}


/*****************************************************************************
*	m228_IDReadBit
*
*	Reads a single bit from the ID PROM of the instrument.
*	The instruments ID PROM is serially accessed
*****************************************************************************/
static Int32 m228_IDReadBit(M228_HANDLE handle,  Bool *value)
{	Int32 error = M228_SUCCESS;
	UInt16 data;

/*  clock low  */ 	
	checkErr(m228_WriteRegister16(handle, M228_ID_ACCESS_REGISTER, 0x4));	
	
/*  clock high  */	
	checkErr(m228_WriteRegister16(handle, M228_ID_ACCESS_REGISTER, 0x6));	
	
/*  Read data  */	
	checkErr(m228_ReadRegister16(handle, M228_ID_ACCESS_REGISTER, &data));
	*value = ( Bool)data;

Error:	
	return error;
}


/*****************************************************************************
*	m228_SetRandomDataAddressPointer
*
******************************************************************************/
Int32 m228_SetRandomDataAddressPointer (M228_HANDLE handle, UInt32 address)
{	Int32 error = M228_SUCCESS;
	
/*  Check parameters  */
	checkParamUnsignedRange(address, 0x3FFFFFF, 2);
	
	checkErr(m228_WriteRegister16(handle, M228_RANDOM_DATA_ADDRESS_REGISTER_LOW, address & 0xFFFF));
	checkErr(m228_WriteRegister16(handle, M228_RANDOM_DATA_ADDRESS_REGISTER_HIGH, (address>>16) & 0x3FF));

Error:		
	return error;
}


/*****************************************************************************
*	m228_GetRandomDataAddressPointer
*
******************************************************************************/
Int32 m228_GetRandomDataAddressPointer (M228_HANDLE handle, UInt32 *address)
{	Int32 error = M228_SUCCESS;
	UInt16 data;
	
/*  Check parameters  */
	checkParamPointer(address, 2);
	
	checkErr(m228_ReadRegister16(handle, M228_RANDOM_DATA_ADDRESS_REGISTER_HIGH, &data));
	*address = (data << 16) & 0x3FF0000;
		
	checkErr(m228_ReadRegister16(handle, M228_RANDOM_DATA_ADDRESS_REGISTER_LOW, &data));
	*address |= data;
		
		
Error:	
	return error;
}


/*****************************************************************************
*	m228_ReadTimeValuePair
*
******************************************************************************/
static Int32 m228_ReadTimeValuePair(M228_HANDLE handle, Int32 interface, Bool width32bit, UInt32 *timestamp, Real64 *value, UInt16 *raw_data)
{	Int32 error = M228_SUCCESS;
	UInt16 data;
	UInt16 reg_address;
	
/*  Check parameters  */
	checkParamRange(width32bit, 0, 0, 2);	// only 16 bit currently supported
	
	if(interface == 0)
		reg_address = M228_FIFO_DATA_REGISTER;
	else
		reg_address = M228_RANDOM_DATA_PORT_REGISTER;

/*	Read the fifo to get the first word  */
	checkErr(m228_ReadRegister16(handle, reg_address, &data));
	
/*  If data valid bit is set then read the fifo 2 more times  */	
	if (data & M228_DATA_VALID_BIT_MASK)  //Check valid bit
	{	if(raw_data != NULL)
			*raw_data = (data & M228_DATA_14_BIT_MASK);
		
		if(timestamp != NULL) // if data only storage mode
		{	checkErr(m228_ReadRegister16(handle, reg_address, &data));
			*timestamp = (data << 16) & 0xFFFF0000;
			checkErr(m228_ReadRegister16(handle, reg_address, &data));
			*timestamp |= data;
		}
		if(value != NULL)
			/* Calculate the real value  */		
			checkErr(m228_ConvertToVoltage(handle, &data, value, 1, 1, 0, 0, 0));
	}
	else
	{	error = M228_ERROR_FIFO_EMPTY;
		goto Error;
	}

Error:
	return error;
}


/*****************************************************************************
*	m228_ReadRandomDataPort
*
******************************************************************************/
Int32 m228_ReadRandomDataPort (M228_HANDLE handle, Bool width32bit, UInt32 *timeStamp, Real64 *value, UInt16 *rawData)
{	Int32 error = M228_SUCCESS;
	
/*  Check parameters  */
	checkParamRange(width32bit, 0, 0, 2);	// only 16 bit currently supported

	checkErr(m228_ReadTimeValuePair(handle, 1, width32bit, timeStamp, value, rawData));

Error:
	return error;
}

 
/*****************************************************************************
*	m228_CalMessage
*
******************************************************************************/
static void m228_CalMessage(void (*user_CalMessage)(Int8 *), Int8 *msg, ...)
{	va_list argp; 
	Int8 string[M228_MAX_CAL_MESSAGE_LENGTH];
	
	if(user_CalMessage != NULL)
	{	va_start(argp, msg);
		vsprintf (string, msg, argp);
		user_CalMessage(string);
		va_end(argp);
	}
}


/*****************************************************************************
*	m228_FindCalibrationValue
*
******************************************************************************/
static Int32 m228_FindCalibrationValue(M228_HANDLE handle, Bool type, Real64 exp, Real64 max, Real64 min, UInt8 *value, Bool *abort, void (*user_CalMessage)(Int8 *))
{	Int32 error = M228_SUCCESS;
 	Real64 reading;
	Int32 num_tries, max_tries;
	UInt8 upper, lower;
  Int32 inc;
	UInt8 fullscale_val = 0x80, offset_val = 0x80;


	m228_CalMessage(user_CalMessage,"Limits: %f exp, %f max, %f min\n", exp, max, min);

	checkErr(m228_QueryCalibrationFullScaleAdjust (handle, &fullscale_val));
	checkErr(m228_QueryCalibrationOffsetAdjust (handle, &offset_val));
	
	if(type == 0)
	{	offset_val = 0x80;
		*value = 0x80;
		checkErr(m228_CalibrationOffsetAdjust (handle, *value));
	}
	else
	{	fullscale_val = 0x80;
		*value = 0x80;
		checkErr(m228_CalibrationFullScaleAdjust (handle, *value));
	}
	m228_CalMessage(user_CalMessage,"Trying FS = %x, OS = %x ... ", (type==0)?fullscale_val:*value, (type==0)?*value:offset_val);
	apis_delay(1500);			//1.5 seconds

	if(*abort == M228_TRUE)
	{	error = M228_ERROR_CALIBRATION_ABORTED;
		goto Error;
	}

	// find correct offset value using a divide and conquer method
	checkErr(m228_CurrentValue (handle, &reading, NULL));
	m228_CalMessage(user_CalMessage,"%fV\n", reading);
	num_tries = 0; max_tries = 40;
	upper = 255; lower = 0;
	while (((reading > max) || (reading < min)) && (num_tries < max_tries) && (*abort == M228_FALSE))
	{	if(reading > exp)
		{ upper = *value;
			inc = (*value - lower) / 2;
			if (inc == 0) 
				inc = 1;
			*value = *value - inc;
			if(type == 0)
				checkErr(m228_CalibrationOffsetAdjust (handle, *value));
			else
				checkErr(m228_CalibrationFullScaleAdjust (handle, *value));
		}
		else
		{ lower = *value;
			inc = (upper - *value) / 2;
			if (inc == 0) 
				inc = 1;
			*value = *value + inc;
			if(type == 0)
				checkErr(m228_CalibrationOffsetAdjust (handle, *value));
			else
				checkErr(m228_CalibrationFullScaleAdjust (handle, *value));
		}
		num_tries++;
		m228_CalMessage(user_CalMessage,"Trying FS = %x, OS = %x ... ", (type==0)?fullscale_val:*value, (type==0)?*value:offset_val);
		apis_delay(1500);
		checkErr(m228_CurrentValue (handle, &reading, NULL));
		m228_CalMessage(user_CalMessage,"%fV\n", reading);
	}

	if(*abort == M228_TRUE)
	{	error = M228_ERROR_CALIBRATION_ABORTED;
		goto Error;
	}

	if (num_tries >= max_tries)
	{	m228_CalMessage(user_CalMessage, "ERROR: Calibration %s value could not be found after %d tries.\n", (type==0)?"Offset":"Fullscale", num_tries);
		error = M228_ERROR_CALIBRATION;
	}

Error:
	return error;
}


/*****************************************************************************
*	m228_SetAdjustedPeripheralVoltage
*
******************************************************************************/
static Int32 m228_SetAdjustedPeripheralVoltage(M228_HANDLE handle, UInt8 offset_cal_val, UInt8 fullscale_cal_val, Real64 voltage)
{	Int32 error = M228_SUCCESS;
	Real64 corrected_voltage, volts_not_adjusted, PVAR_offset_cal;
	UInt8 saved_offset_cal_val, saved_fullscale_cal_val;
	
 	checkErr(m228_QueryCalibrationFullScaleAdjust (handle, &saved_fullscale_cal_val));
	checkErr(m228_QueryCalibrationOffsetAdjust (handle, &saved_offset_cal_val));
 	
	checkErr(m228_CalibrationFullScaleAdjust (handle, fullscale_cal_val));
	checkErr(m228_CalibrationOffsetAdjust (handle, offset_cal_val));

	checkErr(m228_ConfigureReferenceInjection (handle, 1, 5));
	checkErr(m228_SetPeripheralVoltage (handle, voltage));
	apis_delay(200);
	checkErr(m228_CurrentValue (handle, &volts_not_adjusted, NULL));
	PVAR_offset_cal = volts_not_adjusted - voltage;
	corrected_voltage = voltage - PVAR_offset_cal;
	if (corrected_voltage > 10.0) 
		corrected_voltage = 10.0;
	if (corrected_voltage < -10.0) 
		corrected_voltage = -10.0;
	checkErr(m228_SetPeripheralVoltage (handle, corrected_voltage));
	
Error:
 	m228_CalibrationFullScaleAdjust (handle, saved_fullscale_cal_val);
	m228_CalibrationOffsetAdjust (handle, saved_offset_cal_val);
	return error;
}


/*****************************************************************************
*	m228_PerformUnityCalibration
*
******************************************************************************/
static Int32 m228_PerformUnityCalibration(M228_HANDLE handle, UInt8 *offset_val, UInt8 *fullscale_val, M228_CALIBRATION_LIMITS *limits, Bool *abort, void (*user_CalMessage)(Int8 *))
{	Int32 error = M228_SUCCESS;
	UInt8 saved_offset_cal_val, saved_fullscale_cal_val;

 	checkErr(m228_QueryCalibrationFullScaleAdjust (handle, &saved_fullscale_cal_val));
	checkErr(m228_QueryCalibrationOffsetAdjust (handle, &saved_offset_cal_val));

	checkErr(m228_Run (handle, 0, 0));											// disable timestamping and conversion
	checkErr(m228_ResetFifo (handle));
	checkErr(m228_ResetTimeStamp (handle));
	checkErr(m228_DataStorageMode (handle, 3));							// no storage mode
	checkErr(m228_SimulationMode (handle, 0, 0));						// normal input mode
	checkErr(m228_ConfigureSampleClock (handle, 0, 1));			// set sample rate to 500KHz

	checkErr(m228_ConfigureAnalogFrontEnd (handle, 1, 0, 0));		
	checkErr(m228_ConfigureWarpMode(handle, 0));	 																// warp mode = 0 
	checkErr(m228_ConfigureFilter (handle, 0, 0));					
	checkErr(m228_ConfigureFunctionSource (handle, M228_A2D_CONV_ENABLE_FUNCTION, M228_SOFTWARE_REGISTER, 0));

	checkErr(m228_Run (handle, 0, 1));											// conversion enable

	if(*abort == M228_TRUE)
	{	error = M228_ERROR_CALIBRATION_ABORTED;
		goto Error;
	}

/*  Calibrate with divider=0, gain=0, bypass, etc. so that we can in-turn calibrate the peripheral voltage reference  */
	m228_CalMessage(user_CalMessage,"Unity Gain Calibration: used to calibrate the peripheral voltage source.\n");
	*fullscale_val = 0x80;
	*offset_val = 0x80;
	checkErr(m228_CalibrationFullScaleAdjust (handle, *fullscale_val));
	checkErr(m228_CalibrationOffsetAdjust (handle, *offset_val));
	checkErr(m228_ConfigureReferenceInjection (handle, 1, 0));
	apis_delay(1500);   //1.5 seconds
	error = m228_FindCalibrationValue(handle, 0, limits->offset_exp, limits->offset_max, limits->offset_min, offset_val, abort, user_CalMessage);
	if(error < M228_SUCCESS)
		goto Error;
	checkErr(m228_ConfigureReferenceInjection (handle, 1, 3));
	error = m228_FindCalibrationValue(handle, 1, limits->fullscale_exp, limits->fullscale_max, limits->fullscale_min, fullscale_val, abort, user_CalMessage);
	if(error < M228_SUCCESS)
		goto Error;
	else
	{	// increase Fullscale value by 1 to account for reference being +10V
		(*fullscale_val)++;
		checkErr(m228_CalibrationFullScaleAdjust (handle, *fullscale_val));
	}
		
Error:
 	m228_CalibrationFullScaleAdjust (handle, saved_fullscale_cal_val);
	m228_CalibrationOffsetAdjust (handle, saved_offset_cal_val);
	return error;
}
 

/*****************************************************************************
*	m228_CalibrateToInternalReference
*
******************************************************************************/
Int32 m228_CalibrateToInternalReference(M228_HANDLE handle, Bool offset_only, Bool divider, UInt8 fgain, UInt8 bgain, UInt8 filter_type, UInt8 filter_cutoff, M228_CALIBRATION_LIMITS *limits, Bool *abort, void (*user_CalMessage)(Int8 *))
{	Int32 error = M228_SUCCESS;
	Real64 offset_voltage=0, fullscale_voltage=0;
	Real64 total_gain;
	UInt8 fullscale_val, offset_val;
	UInt8 unity_fullscale_val=0, unity_offset_val=0;
	Bool unity_calibration=0;
	Int32 num_tries;
	UInt8 save_fullscale=0x80;
	Int32 i;
	Real64 limit_swap;
	M228_CALIBRATION_LIMITS unity_limits;

/*  Check parameters  */
	checkParamRange(offset_only, 0, 1, 2);
	checkParamRange(divider, 0, 1, 3);
	checkParamRange(fgain, 0, 3, 4);
	checkParamRange(bgain, 0, 7, 5);
	checkParamRange(filter_type, 0, 2, 6);
	checkParamRange(filter_cutoff, 0, 8, 7);
	checkParamPointer(limits, 8);
	checkParamPointer(abort, 9);
	
	/*  Initialize unity calibration limits used when calibrating the peripheral variable voltage reference */
	unity_limits.cal_type = M228_UNITY_CAL_LIMITS_TYPE; 					
	unity_limits.offset_exp = M228_UNITY_CAL_LIMITS_OFFSET_EXP;
	unity_limits.offset_max = M228_UNITY_CAL_LIMITS_OFFSET_MAX;
	unity_limits.offset_min = M228_UNITY_CAL_LIMITS_OFFSET_MIN;
	unity_limits.fullscale_exp = M228_UNITY_CAL_LIMITS_FULLSCALE_EXP;
	unity_limits.fullscale_max = M228_UNITY_CAL_LIMITS_FULLSCALE_MAX;
	unity_limits.fullscale_min = M228_UNITY_CAL_LIMITS_FULLSCALE_MIN;
	
/*  Determin total gain through the analog input path */
	total_gain = bgain_multiplier[bgain] * fgain_multiplier[fgain];
	if(divider == 0)
		total_gain /= 10.0;
	
/*  Calculate fullscale voltage level using gain */	
	if(total_gain != 0)
		fullscale_voltage = 10.0 / total_gain;
	else
		fullscale_voltage = 0;
		
/*  Adjust limits to account for gain  */	
	limits->fullscale_exp = limits->fullscale_exp/total_gain;
	limits->fullscale_max = limits->fullscale_max/total_gain;
	limits->fullscale_min = limits->fullscale_min/total_gain;
	
	if(limits->cal_type == 0) 						//0=unipolar positive, 1=bipolar, 2=unipolar negative
		offset_voltage = 0.0;
	else if (limits->cal_type == 2)			
	{	offset_voltage = 0.0;
		fullscale_voltage *= -1.0;
	}
	else		
		offset_voltage = 0.0;
	
	m228_CalMessage(user_CalMessage, "Internal Reference Calibration \n");
	m228_CalMessage(user_CalMessage, "Calibrating Offset at %fV\n", offset_voltage);
	m228_CalMessage(user_CalMessage, "and Full Scale at %fV\n", fullscale_voltage);
	
	if(*abort == M228_TRUE)
	{	error = M228_ERROR_CALIBRATION_ABORTED;
		goto Error;
	}
		
	fullscale_val = 0x80;
	offset_val = 0x80;
	checkErr(m228_CalibrationFullScaleAdjust (handle, fullscale_val));
	checkErr(m228_CalibrationOffsetAdjust (handle, offset_val));

	// OFFSET Calibration
	m228_CalMessage(user_CalMessage,"Offset Calibration\n");
	if(offset_voltage == 0.0)
		checkErr(m228_ConfigureReferenceInjection (handle, 1, 0));
	else if(offset_voltage == 10.0)
		checkErr(m228_ConfigureReferenceInjection (handle, 1, 3));
	else if (offset_voltage == -10.0)	// -10V
		checkErr(m228_ConfigureReferenceInjection (handle, 1, 4));
	else if (offset_voltage == 0.5)	// -10V
		checkErr(m228_ConfigureReferenceInjection (handle, 1, 1));
	else if (offset_voltage == -0.5)	// -10V
		checkErr(m228_ConfigureReferenceInjection (handle, 1, 2));
	else
	{	checkErr(m228_PerformUnityCalibration(handle, &unity_offset_val, &unity_fullscale_val, &unity_limits, abort, user_CalMessage));
		unity_calibration = 1;
		checkErr(m228_SetAdjustedPeripheralVoltage(handle, unity_offset_val, unity_fullscale_val, offset_voltage));
	}
	
	
	if(*abort == M228_TRUE)
	{	error = M228_ERROR_CALIBRATION_ABORTED;
		goto Error;
	}

	checkErr(m228_Run (handle, 0, 0));											// disable timestamping and conversion
	checkErr(m228_ResetFifo (handle));
	checkErr(m228_ResetTimeStamp (handle));
	checkErr(m228_DataStorageMode (handle, 3));							// no storage mode
	checkErr(m228_SimulationMode (handle, 0, 0));						// normal input mode
	checkErr(m228_ConfigureSampleClock (handle, 0, 1));			// set sample rate to 500KHz

	checkErr(m228_ConfigureAnalogFrontEnd (handle, divider, fgain, bgain));		
	checkErr(m228_ConfigureWarpMode(handle, 0));	 																// warp mode = 0 
	checkErr(m228_ConfigureFilter (handle, filter_cutoff, filter_type));					
	checkErr(m228_ConfigureFunctionSource (handle, M228_A2D_CONV_ENABLE_FUNCTION, M228_SOFTWARE_REGISTER, 0));
	checkErr(m228_Run (handle, 0, 1));											// conversion enable
	
	apis_delay(1500);   //1.5 seconds
	
	if(*abort == M228_TRUE)
	{	error = M228_ERROR_CALIBRATION_ABORTED;
		goto Error;
	}

	error = m228_FindCalibrationValue(handle, 0, limits->offset_exp, limits->offset_max, limits->offset_min, &offset_val, abort, user_CalMessage);
	
	// FULLSCALE Calibration
	if ((error == M228_SUCCESS) && (offset_only == 0))
	{ m228_CalMessage(user_CalMessage,"Fullscale Calibration\n");
		
		if(limits->cal_type == 1) 						//0=unipolar positive, 1=bipolar, 2=unipolar negative
			num_tries = 2;
		else
			num_tries = 1;
			
		for(i=0; ((i<num_tries) && (*abort==M228_FALSE)); i++)
		{	if(i>0)
			{	fullscale_voltage *= -1.0;
				limits->fullscale_exp *= -1.0;
				limit_swap = limits->fullscale_max;
				limits->fullscale_max = limits->fullscale_min * -1.0;
				limits->fullscale_min = limit_swap  * -1.0;
			}
				
			if(fullscale_voltage == 10.0)
				checkErr(m228_ConfigureReferenceInjection (handle, 1, 3));
			else if (fullscale_voltage == -10.0)	// -10V
				checkErr(m228_ConfigureReferenceInjection (handle, 1, 4));
			else if (fullscale_voltage == 0.5)	// -10V
				checkErr(m228_ConfigureReferenceInjection (handle, 1, 1));
			else if (fullscale_voltage == -0.5)	// -10V
				checkErr(m228_ConfigureReferenceInjection (handle, 1, 2));
			else
			{	if(unity_calibration == 0)
				{	checkErr(m228_PerformUnityCalibration(handle, &unity_offset_val, &unity_fullscale_val, &unity_limits, abort, user_CalMessage));
					unity_calibration = 1;
				}
				checkErr(m228_SetAdjustedPeripheralVoltage(handle, unity_offset_val, unity_fullscale_val, fullscale_voltage));
			}
		
			if(*abort == M228_TRUE)
			{	error = M228_ERROR_CALIBRATION_ABORTED;
				goto Error;
			}
			
			checkErr(m228_Run (handle, 0, 0));											// disable timestamping and conversion
			checkErr(m228_ResetFifo (handle));
			checkErr(m228_ResetTimeStamp (handle));
			checkErr(m228_DataStorageMode (handle, 3));							// no storage mode
			checkErr(m228_SimulationMode (handle, 0, 0));						// normal input mode
			checkErr(m228_ConfigureSampleClock (handle, 0, 1));			// set sample rate to 500KHz

			checkErr(m228_ConfigureAnalogFrontEnd (handle, divider, fgain, bgain));		
			checkErr(m228_ConfigureWarpMode(handle, 0));	 																// warp mode = 0 
			checkErr(m228_ConfigureFilter (handle, filter_cutoff, filter_type));					
			checkErr(m228_ConfigureFunctionSource (handle, M228_A2D_CONV_ENABLE_FUNCTION, M228_SOFTWARE_REGISTER, 0));
			checkErr(m228_Run (handle, 0, 1));											// conversion enable
	
			apis_delay(1500);   //1.5 seconds

			if(*abort == M228_TRUE)
			{	error = M228_ERROR_CALIBRATION_ABORTED;
				goto Error;
			}
			error = m228_FindCalibrationValue(handle, 1, limits->fullscale_exp, limits->fullscale_max, limits->fullscale_min, &fullscale_val, abort, user_CalMessage);

			if(error == M228_SUCCESS)
			{	// increase Fullscale value by 1 to account for reference being +10V
				fullscale_val++;
				if((limits->cal_type == 1) && (i == 0)) 						//cal_type: 0=unipolar positive, 1=bipolar, 2=unipolar negative
					save_fullscale = fullscale_val;
				else if(limits->cal_type == 1)
					fullscale_val = (fullscale_val + save_fullscale)/2;
			}
		}

		if(error == M228_SUCCESS)
		{	
			checkErr(m228_CalibrationFullScaleAdjust (handle, fullscale_val));
			m228_CalMessage(user_CalMessage,"Final Settings FS = %x, OS = %x\n", fullscale_val, offset_val);
		}
	}
	
Error:
	return error;
}



/*****************************************************************************
*	Modify the functions below this point to use the driver in an 
*	environment not supported by an APIS platform support module. All functions
*	above this point should not have to be modified.
*
*	In general, any call to an APIS function (apis_func()) will need to be
*	modified to communicate with the instrument or operating system being used.
*
******************************************************************************/
/*****************************************************************************
*	m228_Init
*
*	Opens a session to the instrument specified by the index parameter.  
*	If IDQuery is true then this function verifies that the instrument is a 
*	m228.  If reset is true then this function resets the instrument to it's 
*	default state.
******************************************************************************/
Int32 m228_Init (Int32 instr,  Bool IDQuery,  Bool reset, M228_HANDLE *handle)
{	Int32 error = M228_SUCCESS;
	Int32 warning = M228_SUCCESS;
  UInt16 value;

/*  Check parameters  */
	checkParamRange(IDQuery, 0, 1, 2);
	checkParamRange(reset, 0, 1, 3);
	checkParamPointer(handle, 4);
	
/*  Allocate the data structure  */
	(*handle) = 0;
	(*handle) = malloc(sizeof(M228_INSTR));
	if((*handle) == NULL)
	{	error = M228_ERROR_OUT_OF_MEMORY;
		goto Error;
	}

	(*handle)->apis_handle = 0;
	apisCheckErrInit(apis_open((APIS_PATH)instr, &(*handle)->apis_handle));
	
/* Perform ID Query */
	if(IDQuery == M228_TRUE)
	{	checkErr(m228_GetIdentWord (*handle, 17, &value));
	  
		value &= 0x0FFF;
		if (value != (UInt16)M228_VXI_MANF_ID) 
		{	error = M228_ERROR_FAIL_ID_QUERY;
 			goto Error;
	  }

		checkErr(m228_GetIdentWord (*handle, 18, &value));
		
		value &= 0x0FFF;
		if (value != (UInt16)M228_VXI_MODEL_CODE) 
		{	error = M228_ERROR_FAIL_ID_QUERY;
			goto Error;
		} 
	}
/*-- End of ID Query --*/


/* Reset Instrument */
  if (reset == M228_TRUE)
  {	checkErr(m228_Reset (*handle));
  }

Error:
		
	if((error == M228_SUCCESS) && (warning != M228_SUCCESS))		// if no errors then return the warnings		
		error = warning;

	return error;
}


/*****************************************************************************
*	m228_Close
*
*	Close the VISA session to the instrument.
******************************************************************************/
Int32 m228_Close (M228_HANDLE handle)
{	Int32 error = M228_SUCCESS;
	if(handle->apis_handle != NULL)
		apisCheckErr(apis_close(handle->apis_handle));
	free(handle);
Error:
	return error;
}
	


/*****************************************************************************
*	m228_Reset
*	
*	Reset the instrument to it's default state
******************************************************************************/
Int32 m228_Reset (M228_HANDLE handle)
{	Int32 error = M228_SUCCESS;
	apisCheckErr(apis_reset(handle->apis_handle));
Error:
	return error;
}


/*****************************************************************************
*	m228_ReadRegister16
*
*	Read the specified offset from the module's I/O space and returns
*	the result.
******************************************************************************/
Int32 m228_ReadRegister16 (M228_HANDLE handle, UInt16 offset, UInt16 *value)
{	Int32 error = M228_SUCCESS;
	
/*  Check parameters  */
	checkParamPointer(value, 3);

	apisCheckErr(apis_read(handle->apis_handle, 2, offset, value));
Error:
	return error;
}


/*****************************************************************************
*	m228_ReadRegister32
*
*	Read the specified offset from the module's I/O space and returns
*	the result.
******************************************************************************/
Int32 m228_ReadRegister32 (M228_HANDLE handle, UInt16 offset, UInt32 *value)
{	Int32 error = M228_SUCCESS;
	
/*  Check parameters  */
	checkParamPointer(value, 3);

	apisCheckErr(apis_read(handle->apis_handle, 4, offset, value));
Error:
	return error;
}



/*****************************************************************************
*	m228_WriteRegister16
*
*	Writes to the specified offset in the module's I/O space
******************************************************************************/
Int32 m228_WriteRegister16 (M228_HANDLE handle, UInt16 offset, UInt16 value)
{	Int32 error = M228_SUCCESS;
	
	apisCheckErr(apis_write(handle->apis_handle, 2, offset, value));
Error:
	return error;
}


/*****************************************************************************
*	m228_WriteRegister32
*
*	Writes to the specified offset in the module's I/O space
******************************************************************************/
Int32 m228_WriteRegister32 (M228_HANDLE handle, UInt16 offset, UInt32 value)
{	Int32 error = M228_SUCCESS;
	
	apisCheckErr(apis_write(handle->apis_handle, 4, offset, value));
Error:
	return error;
}


/*****************************************************************************
*	m228_InstallIrq
*	
*	Installs the interrupts handler to be called when an interrupt occurs
*	and configures the carrier to pass interrupts onto the system.
******************************************************************************/
Int32 m228_InstallIrq (M228_HANDLE handle, void *interruptHandler, Int32 vector, Int32 level, Int32 mode, void *var_ptr)
{	Int32 error = M228_SUCCESS;

/*  Check parameters  */
	checkParamPointer(interruptHandler, 2);

	apisCheckErr(apis_irqinstall(handle->apis_handle, interruptHandler, vector, level, mode, var_ptr)); 

Error:
	return error;
}


/*****************************************************************************
*	m228_UninstallIrq
*
*	Un-installs the interrupt handler for the instrument and disables
*	interrupt handling for the instrument
******************************************************************************/
Int32 m228_UninstallIrq (M228_HANDLE handle)
{	Int32 error = M228_SUCCESS;

	apisCheckErr(apis_irqremove(handle->apis_handle));

Error:
	return error;
}


/*****************************************************************************
*	m228_GetApisError
*
*	Returns the saved apis error code
******************************************************************************/
Int32 m228_GetApisError (M228_HANDLE handle, UInt32 *apis_error)
{	Int32 error = M228_SUCCESS;
	*apis_error = handle->apis_error;
	return error;
}



